"""
Test whether the basic stuff of Registry works as intended
"""
import logging
import unittest

import pkg_resources
from collections import OrderedDict

from dhcpkit.ipv6.options import ClientIdOption, ServerIdOption
from dhcpkit.registry import Registry


class TestRegistry(Registry):
    entry_point = 'dhcpkit.tests.registry'


class ElementOccurrenceTestCase(unittest.TestCase):
    def test_registry_loading(self):
        entry_map = pkg_resources.get_entry_map('dhcpkit')

        # Steal the distribution from an existing entry
        dist = entry_map['dhcpkit.ipv6.options']['1'].dist

        entry_map['dhcpkit.tests.registry'] = {
            # Test string ID
            'one': pkg_resources.EntryPoint.parse('one = dhcpkit.ipv6.options:ClientIdOption', dist=dist),

            # Test numerical ID
            '1': pkg_resources.EntryPoint.parse('1 = dhcpkit.ipv6.options:ServerIdOption', dist=dist),
        }

        test_registry = TestRegistry()
        self.assertEqual(len(test_registry), 2)
        self.assertEqual(test_registry['one'], ClientIdOption)
        self.assertEqual(test_registry[1], ServerIdOption)

    def test_duplicate_entries(self):
        entry_map = pkg_resources.get_entry_map('dhcpkit')

        # Steal the distribution from an existing entry
        dist = entry_map['dhcpkit.ipv6.options']['1'].dist

        # Test one ID appearing multiple times (can happen if other packages overwrite our keys)
        entry_map['dhcpkit.tests.registry'] = OrderedDict()
        entry_map['dhcpkit.tests.registry']['1'] = pkg_resources.EntryPoint.parse(
            '1 = dhcpkit.ipv6.options:ClientIdOption', dist=dist)
        entry_map['dhcpkit.tests.registry']['1 '] = pkg_resources.EntryPoint.parse(
            '1 = dhcpkit.ipv6.options:ServerIdOption', dist=dist)

        with self.assertLogs('dhcpkit.registry', logging.WARNING) as cm:
            test_registry = TestRegistry()

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(cm.output[0], '^WARNING:.*:Multiple entry points found for TestRegistry 1')

        self.assertEqual(len(test_registry), 1)
        self.assertEqual(test_registry[1], ClientIdOption)

    def test_bad_entry(self):
        entry_map = pkg_resources.get_entry_map('dhcpkit')

        # Steal the distribution from an existing entry
        dist = entry_map['dhcpkit.ipv6.options']['1'].dist

        entry_map['dhcpkit.tests.registry'] = {
            # Test something that doesn't exist
            'bad': pkg_resources.EntryPoint.parse('bad = dhcpkit.tests.does_not_exist:DummyOption', dist=dist),
        }

        with self.assertLogs('dhcpkit.registry', logging.WARNING) as cm:
            test_registry = TestRegistry()

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(cm.output[0], '^ERROR:.*:Entry point bad .* could not be loaded')

        self.assertEqual(len(test_registry), 0)


if __name__ == '__main__':
    unittest.main()



"""
Handlers to apply to transaction bundles
"""

import abc
import logging

from dhcpkit.common.server.config_elements import ConfigElementFactory
from dhcpkit.ipv6.messages import RelayForwardMessage, RelayReplyMessage
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


class HandlerException(Exception):
    """
    Base class for handler exceptions
    """


class CannotRespondError(HandlerException):
    """
    This exception signals that we cannot reply to this client.
    """


class UseMulticastError(HandlerException):
    """
    This exception signals that a STATUS_USEMULTICAST should be returned to the client.
    """


class Handler:
    """
    Base class for handlers
    """

    def __init__(self):
        """
        The main initialisation will be done in the master process. After initialisation the master process will create
        worker processes using the multiprocessing module. Things that can't be pickled and transmitted to the worker
        processes (think database connections etc) have to be initialised separately. Each worker process will call
        worker_init() to do so. Filters that don't need per-worker initialisation can do everything here in __init__().
        """

    def __str__(self):
        """
        Return a representation of this handler for logging purposes

        :return: A descriptive string
        """
        # Use the class name as default, let subclasses overrule this where it makes sense
        return self.__class__.__name__

    def worker_init(self):
        """
        Separate initialisation that will be called in each worker process that is created. Things that can't be forked
        (think database connections etc) have to be initialised here.
        """

    def analyse_pre(self, bundle: TransactionBundle):
        """
        Analyse the request that came in before handlers can change it.

        :param bundle: The transaction bundle
        """

    def pre(self, bundle: TransactionBundle):
        """
        Pre-process the data in the bundle. Subclasses can update bundle state here or abort processing of the request
        by raising a CannotRespondError.

        :param bundle: The transaction bundle
        """

    def handle(self, bundle: TransactionBundle):
        """
        Handle the data in the bundle. Subclasses should do their main work here.

        :param bundle: The transaction bundle
        """

    def post(self, bundle: TransactionBundle):
        """
        Post-process the data in the bundle. Subclasses can e.g. clean up state. Subclasses assigning addresses should
        check whether the bundle.response is an AdvertiseMessage or a ReplyMessage. The class can change between
        handle() and post() when the server is using rapid-commit.

        :param bundle: The transaction bundle
        """

    def analyse_post(self, bundle: TransactionBundle):
        """
        Analyse the response that is going out after all handlers have been applied.

        :param bundle: The transaction bundle
        """


class RelayHandler(Handler, metaclass=abc.ABCMeta):
    """
    A base class for handlers that work on option in the relay messages chain.
    """

    def handle(self, bundle: TransactionBundle):
        """
        Handle the data in the bundle by checking the relay chain and calling :meth:`handle_relay` for each relay
        message.

        :param bundle: The transaction bundle
        """
        # We need the outgoing chain to be present
        if bundle.outgoing_relay_messages is None:
            logger.error("Cannot process relay chains: outgoing chain not set")
            return

        # Don't try to match between chains of different size
        if len(bundle.incoming_relay_messages) != len(bundle.outgoing_relay_messages):
            logger.error("Cannot process relay chains: chain have different length")
            return

        # Process the relay messages one by one
        for relay_message_in, relay_message_out in zip(bundle.incoming_relay_messages, bundle.outgoing_relay_messages):
            self.handle_relay(bundle, relay_message_in, relay_message_out)

    @abc.abstractmethod
    def handle_relay(self, bundle: TransactionBundle,
                     relay_message_in: RelayForwardMessage, relay_message_out: RelayReplyMessage):
        """
        Handle the options for each relay message pair.

        :param bundle: The transaction bundle
        :param relay_message_in: The incoming relay message
        :param relay_message_out: Thr outgoing relay message
        """


class HandlerFactory(ConfigElementFactory, metaclass=abc.ABCMeta):
    """
    Base class for handler factories
    """



"""
Test the ServerUnicastOption implementation
"""
import unittest
from ipaddress import IPv6Address

from dhcpkit.ipv6.options import ServerUnicastOption
from tests.ipv6.options import test_option


class ServerUnicastOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('000c001020010db8000000000000000000000001')
        self.option_object = ServerUnicastOption(IPv6Address('2001:db8::1'))
        self.parse_option()

    def test_server_address(self):
        self.option.server_address = bytes.fromhex('20010db8000000000000000000000001')
        with self.assertRaisesRegex(ValueError, 'valid IPv6 address'):
            self.option.validate()

        self.option.server_address = IPv6Address('2001:db8::1')
        self.option.validate()

        self.option.server_address = IPv6Address('::')
        with self.assertRaisesRegex(ValueError, 'valid IPv6 address'):
            self.option.validate()

        self.option.server_address = IPv6Address('::1')
        with self.assertRaisesRegex(ValueError, 'valid IPv6 address'):
            self.option.validate()

        self.option.server_address = IPv6Address('ff02::1')
        with self.assertRaisesRegex(ValueError, 'valid IPv6 address'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must have length 16'):
            ServerUnicastOption.parse(bytes.fromhex('000c000f'))

        with self.assertRaisesRegex(ValueError, 'must have length 16'):
            ServerUnicastOption.parse(bytes.fromhex('000c0011'))


if __name__ == '__main__':
    unittest.main()



"""
An object to hold everything related to a request/response transaction
"""
import codecs
import logging
from ipaddress import IPv6Address

from dhcpkit.ipv6.messages import Message, RelayForwardMessage, ClientServerMessage, RelayReplyMessage
from dhcpkit.ipv6.options import Option, ClientIdOption
from dhcpkit.ipv6.utils import split_relay_chain
from typing import Tuple, List, TypeVar, Type, Iterable

logger = logging.getLogger(__name__)

# Typing helpers
SomeOption = TypeVar('SomeOption', bound='Option')


class TransactionBundle:
    """
    A bundle with all data about a transaction. This makes it much easier to pass around multiple pieces of information.

    :type incoming_message: Message
    :type received_over_multicast: bool
    :type request: ClientServerMessage
    :type incoming_relay_messages: List[RelayForwardMessage]
    :type response: Optional[ClientServerMessage]
    :type outgoing_relay_messages: Optional[List[RelayReplyMessage]]
    :type handled_options: List[Option]
    :type marks: Set[str]
    :type handler_data: Dict[Handler, object]
    """

    def __init__(self, incoming_message: Message, received_over_multicast: bool, allow_rapid_commit: bool = False,
                 marks: Iterable[str] = None):

        self.incoming_message = incoming_message
        """The incoming message including the relay chain"""

        self.received_over_multicast = received_over_multicast
        """A flag indicating whether the client used multicast to contact the server"""

        self.allow_unicast = False
        """Allow the client use unicast to contact the server. Set to True by handlers"""

        self.allow_rapid_commit = allow_rapid_commit
        """Allow rapid commit? May be set to True on creation, may be set to False by handlers, not vice versa"""

        # Convenience properties for easy access to the request and chain without having to walk the chain every time
        self.request = None
        """The incoming request without the relay messages"""

        self.incoming_relay_messages = []
        """The chain of relay messages starting with the one closest to the client"""

        self.request, self.incoming_relay_messages = split_relay_chain(incoming_message)

        self.response = None
        """This is where the user puts the response :class:`.ClientServerMessage`"""

        self.outgoing_relay_messages = None
        """This is where the user puts the reply relay chain by calling :meth:`create_outgoing_relay_messages`"""

        # Extra state to track handling of the message
        self.handled_options = []
        """A list of options from the request that have been handled, only applies to IA type options"""

        self.marks = set(marks or [])
        """A set of marks that have been applied to this message"""

        self.handler_data = {}
        """A place for handlers to store data related to this transaction"""

    def __str__(self) -> str:
        client_id = self.request.get_option_of_type(ClientIdOption)
        if client_id:
            duid = codecs.encode(client_id.save(), 'hex').decode('ascii')
        else:
            duid = 'unknown'

        output = "{} from {}".format(type(self.request).__name__, duid)

        interesting_relay_messages = [message for message in self.incoming_relay_messages
                                      if not message.link_address.is_unspecified]
        if interesting_relay_messages:
            output += ' at {} via {}'.format(interesting_relay_messages[0].peer_address,
                                             interesting_relay_messages[0].link_address)
            for relay in interesting_relay_messages[1:]:
                output += ' -> {}'.format(relay.link_address)

        if self.marks:
            output += " with marks '{}'".format("', '".join(self.marks))

        return output

    def mark_handled(self, option: Option):
        """
        Mark the given option as handled. Not all options are specifically handled. This is mostly useful for
        options like IANAOption, IATAOption and IAPDOption.

        :param option: The option to mark as handled
        """
        if option not in self.handled_options:
            self.handled_options.append(option)

    def get_unhandled_options(self, option_types: Type[SomeOption] or Tuple[Type[SomeOption]]) -> List[SomeOption]:
        """
        Get a list of all Options in the request that haven't been marked as handled

        :return: The list of unanswered Options
        """
        # Make a list of requested IANAOptions
        return [option for option in self.request.options
                if isinstance(option, option_types) and option not in self.handled_options]

    def add_mark(self, mark: str):
        """
        Add this mark to the set.

        :param mark: The mark to add
        """
        self.marks.add(mark.strip())

    @property
    def link_address(self) -> IPv6Address:
        """
        Find the link address that identifies where this request is coming from
        """
        # Start with the relay closest to the client and keep looking until a useful address is found
        for relay in self.incoming_relay_messages:
            # Some relays (i.e. LDRA: :rfc:`6221`) don't have a useful link-address
            if not relay.link_address.is_unspecified and \
                    not relay.link_address.is_loopback and \
                    not relay.link_address.is_link_local:
                # This looks useful
                return relay.link_address

        # Nothing useful...
        return IPv6Address('::')

    @property
    def relays(self) -> List[IPv6Address]:
        """
        Get a list of all the relays that this message went through
        """
        return [relay.link_address for relay in self.incoming_relay_messages
                if not relay.link_address.is_unspecified]

    def create_outgoing_relay_messages(self):
        """
        Create a plain chain of RelayReplyMessages for the current response
        """
        if not self.response:
            raise ValueError("Cannot create outgoing relay messages without a response")

        self.outgoing_relay_messages = []
        if not self.incoming_relay_messages:
            return

        outgoing_message = self.incoming_relay_messages[-1].wrap_response(self.response)
        self.outgoing_relay_messages = []
        while isinstance(outgoing_message, RelayReplyMessage):
            self.outgoing_relay_messages.insert(0, outgoing_message)
            outgoing_message = outgoing_message.relayed_message

    @property
    def outgoing_message(self):
        """
        Wrap the response in a relay chain if necessary
        """
        if self.response is None:
            # No response is ok
            return None

        if not self.response.from_server_to_client:
            logger.error("A server should not send {} to a client".format(self.response.__class__.__name__))
            return None

        if self.incoming_relay_messages and not self.outgoing_relay_messages:
            # No outgoing relay messages, but we had incoming relay messages: auto-create a plain relay chain
            self.create_outgoing_relay_messages()

        if self.outgoing_relay_messages:
            # Make sure the right response is in the relay messages (in case someone overwrites :attr:`response` without
            # updating the contents of the relay messages as well
            self.outgoing_relay_messages[0].relayed_message = self.response

            # Send the relay messages
            return self.outgoing_relay_messages[-1]
        else:
            # Send the plain response
            return self.response



"""
Configuration elements for the dns option handlers
"""
from dhcpkit.ipv6.server.extensions.dns import DomainSearchListOptionHandler, RecursiveNameServersOptionHandler
from dhcpkit.ipv6.server.handlers import HandlerFactory


class RecursiveNameServersOptionHandlerFactory(HandlerFactory):
    """
    Create the handler for recursive name servers.
    """

    def create(self) -> RecursiveNameServersOptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """
        return RecursiveNameServersOptionHandler(dns_servers=self.addresses, always_send=self.always_send)


class DomainSearchListOptionHandlerFactory(HandlerFactory):
    """
    Create the handler for the domain search list.
    """

    def create(self) -> DomainSearchListOptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """
        return DomainSearchListOptionHandler(search_list=self.domain_names, always_send=self.always_send)



import unittest

from dhcpkit.common.server.config_datatypes import domain_name


class DomainNameTestCase(unittest.TestCase):
    def test_valid(self):
        valid_domain_names = [
            ('steffann.nl', 'steffann.nl'),
            ('Steffann.Nl', 'steffann.nl'),
            ('STEFFANN.NL', 'steffann.nl'),
            ('10ww.steffann.nl', '10ww.steffann.nl'),
            ('local', 'local'),
            ('LOCAL', 'local'),
        ]

        for test, output in valid_domain_names:
            with self.subTest(test=test):
                self.assertEqual(domain_name(test), output)

    def test_name_too_long(self):
        oversized_domain_name = '10ww.' \
                                'steffann-steffann-steffann-steffann-steffann-steffann-steffann.' \
                                'steffann-steffann-steffann-steffann-steffann-steffann-steffann.' \
                                'steffann-steffann-steffann-steffann-steffann-steffann-steffann.' \
                                'steffann-steffann-steffann-steffann-steffann-steffann-steffann.' \
                                'nl'
        with self.assertRaisesRegex(ValueError, 'must be 255 characters'):
            domain_name(oversized_domain_name)

    def test_label_too_long(self):
        oversized_label_name = '10ww.steffann-steffann-steffann-steffann-steffann-steffann-steffann-steffann.nl'
        with self.assertRaisesRegex(ValueError, 'must be 1 to 63 characters'):
            domain_name(oversized_label_name)


if __name__ == '__main__':
    unittest.main()



"""
The basic configuration objects for logging
"""
import logging
import logging.handlers
import os
import sys

from ZConfig.datatypes import SocketAddress, existing_dirpath
from ZConfig.matcher import SectionValue

from dhcpkit.common.server.config_elements import ConfigElementFactory, ConfigSection
from dhcpkit.common.server.logging.verbosity import set_verbosity_logger


class Logging(ConfigSection):
    """
    Class managing the configured logging handlers.
    """

    def validate_config_section(self):
        """
        Check for duplicate handlers
        """
        # Check that we don't have multiple console loggers
        have_console = False
        for handler_factory in self.handlers:
            if isinstance(handler_factory, ConsoleHandlerFactory):
                if have_console:
                    raise ValueError("You cannot log to the console multiple times")
                have_console = True

    def configure(self, logger: logging.Logger, verbosity: int = 0) -> int:
        """
        Add all configured handlers to the supplied logger. If verbosity > 0 then make sure we have a console logger
        and force the level of the console logger based on the verbosity.

        :param logger: The logger to add the handlers to
        :param verbosity: The verbosity level given as command line argument
        :return: The lowest log level that is going to be handled
        """
        # Remove any previously configured loggers, in case we are re-configuring
        # We are deleting, so copy the list first
        for handler in list(logger.handlers):
            logger.removeHandler(handler)

        # Add the handlers, keeping track of console loggers and saving the one with the "best" level.
        console = None
        for handler_factory in self.handlers:
            handler = handler_factory()
            logger.addHandler(handler)

            if isinstance(handler_factory, ConsoleHandlerFactory):
                console = handler

        # Set according to verbosity
        set_verbosity_logger(logger, verbosity, console)

        # Find the lowest log level
        lowest_level = logging.CRITICAL
        for handler in logger.handlers:
            if handler.level < lowest_level:
                lowest_level = handler.level

        # Return the lowest log level we want, so that we can filter lower priority messages earlier (where appropriate)
        return lowest_level


class ConsoleHandlerFactory(ConfigElementFactory):
    """
    Factory for a logging handler that logs to the console, optionally in colour.
    """

    def __init__(self, section: SectionValue):
        self.colorlog = None
        super().__init__(section)

    def validate_config_section(self):
        """
        Validate the colorlog setting
        """
        # Try loading colorlog
        try:
            if self.color is False:
                # Explicitly disabled
                colorlog = None
            else:
                # noinspection PyPackageRequirements
                import colorlog
        except ImportError:
            if self.color is True:
                # Explicitly enabled, and failed
                raise ValueError("Colored logging turned on but the 'colorlog' package is not installed")

            colorlog = None

        self.colorlog = colorlog

    def create(self) -> logging.StreamHandler:
        """
        Create a console handler

        :return: The logging handler
        """
        handler = logging.StreamHandler(sys.stderr)
        if self.colorlog and sys.stderr.isatty():
            formatter = self.colorlog.ColoredFormatter('{yellow}{asctime}{reset} '
                                                       '{purple}{processName}:{reset} '
                                                       '[{log_color}{levelname}{reset}] '
                                                       '{white}{message}{reset}',
                                                       style='{')
        else:
            formatter = logging.Formatter('{asctime} {processName}: [{levelname}] {message}',
                                          style='{')

        handler.setFormatter(formatter)
        handler.setLevel(self.level)

        return handler


class FileHandlerFactory(ConfigElementFactory):
    """
    Factory for a logging handler that logs to a file, optionally rotating it.
    """

    name_datatype = staticmethod(existing_dirpath)

    def __init__(self, section: SectionValue):
        super().__init__(section)

    def validate_config_section(self):
        """
        Validate if the combination of settings is valid
        """
        # Size-based rotation and specifying a size go together
        if self.size and self.rotate != 'SIZE':
            raise ValueError("You can only specify a size when rotating based on size")
        elif not self.size and self.rotate == 'SIZE':
            raise ValueError("When rotating based on size you must specify a size")

        # Rotation and keeping old logs go together
        if self.keep and not self.rotate:
            raise ValueError("You can only specify how many log files to keep when rotation is enabled")
        elif not self.keep and self.rotate:
            raise ValueError("You must specify how many log files to keep when rotation is enabled")

    def create(self) -> logging.StreamHandler:
        """
        Create a console handler

        :return: The logging handler
        """
        if self.section.rotate == 'SIZE':
            # Rotate based on file size
            handler = logging.handlers.RotatingFileHandler(filename=self.name,
                                                           maxBytes=self.section.size,
                                                           backupCount=self.section.keep)
        elif self.section.rotate is not None:
            # Rotate on time
            handler = logging.handlers.TimedRotatingFileHandler(filename=self.name,
                                                                when=self.section.rotate,
                                                                backupCount=self.section.keep)
        else:
            # No rotation specified, used a WatchedFileHandler so that external rotation works
            handler = logging.handlers.WatchedFileHandler(filename=self.section.path)

        formatter = logging.Formatter('{asctime} {processName}: [{levelname}] {message}',
                                      style='{')

        handler.setLevel(self.section.level)
        handler.setFormatter(formatter)
        return handler


class SysLogHandlerFactory(ConfigElementFactory):
    """
    Factory for a logging handler that logs to syslog.
    """
    default_destinations = (
        '/dev/log',
        '/var/run/syslog',
        'localhost:514',
    )

    name_datatype = staticmethod(SocketAddress)

    def clean_config_section(self):
        """
        Fill in the name automatically if not given
        """
        # The name is the destination
        if not self.name:
            # Fallback in case no destination is specified
            for destination in self.default_destinations:
                if destination.startswith('/'):
                    if os.path.exists(destination):
                        # Destination is a path, check if it exists
                        self.name = SocketAddress(destination)
                        break
                else:
                    # Not a path, just assume it's ok
                    self.name = SocketAddress(destination)
                    break

    def create(self) -> logging.handlers.SysLogHandler:
        """
        Create a syslog handler

        :return: The logging handler
        """
        handler = logging.handlers.SysLogHandler(address=self.name.address,
                                                 facility=self.section.facility,
                                                 socktype=self.section.protocol)
        handler.setLevel(self.section.level)
        return handler



"""
Extensions to Pygments to correctly parse DHCPKit config files
"""
import re

from pygments.lexer import RegexLexer, bygroups
from pygments.token import Text, Comment, Name, String, Keyword


class DHCPKitConfLexer(RegexLexer):
    """
    Lexer for configuration files following the DHCPKit config file format.
    """

    name = 'DHCPKitConf'
    aliases = ['dhcpkitconf', 'dhcpkit']
    flags = re.MULTILINE | re.IGNORECASE

    tokens = {
        'root': [
            (r'\s+', Text),
            (r'(#.*?)$', Comment),
            (r'(<[^\s>]+)(?:(\s+)(.*?))?(>)',
             bygroups(Name.Tag, Text, String, Name.Tag)),
            (r'([a-z][\w-]*)(\s+)',
             bygroups(Name.Builtin, Text), 'value'),
            (r'\.+', Text),
        ],
        'value': [
            (r'\\\n', Text),
            (r'$', Text, '#pop'),
            (r'\\', Text),
            (r'[^\S\n]+', Text),
            (r'/([a-z0-9][\w./-]+)', String.Other),
            (r'(on|off|yes|no|true|false|'
             r'critical|error|warn|warning|info|debug-packets|debug-handling|debug|notset|'
             r'authpriv|auth|cron|daemon|ftp|kern|lpr|mail|news|security|syslog|uucp|local[0-7]|'
             r'udp|dgram|tcp|stream|'
             r'hourly|hour|daily|day|weekly|week|size|'
             r'user|group)\b', Keyword),
            (r'"([^"\\]*(?:\\.[^"\\]*)*)"', String.Double),
            (r'[^\s"\\]+', Text)
        ],
    }



"""
All the unit tests go here
"""
import copy
from unittest import mock


class DeepCopyMagicMock(mock.MagicMock):
    """
    A magic mock class that deep-copies the method arguments to check the state of mutable objects at call time
    """
    def _mock_call(self, *args, **kwargs):
        return super()._mock_call(*copy.deepcopy(args), **copy.deepcopy(kwargs))



"""
A simple handler that tells the client to use multicast to reach this server.
"""
import logging
from ipaddress import IPv6Address

from dhcpkit.ipv6.options import ServerUnicastOption
from dhcpkit.ipv6.server.handlers import Handler, HandlerFactory, UseMulticastError
from dhcpkit.ipv6.server.handlers.basic import SimpleOptionHandler
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


class RejectUnwantedUnicastHandler(Handler):
    """
    A simple handler that tells the client to use multicast to reach this server.
    """

    def pre(self, bundle: TransactionBundle):
        """
        Reject unicast messages

        :param bundle: The transaction bundle
        """
        if not bundle.allow_unicast and not bundle.received_over_multicast:
            logging.info("Rejecting unicast {}".format(bundle))
            raise UseMulticastError("This server does not support unicast requests")


class ServerUnicastOptionHandler(SimpleOptionHandler):
    """
    A simple handler that tells the client that it may use unicast to contact this server.
    """

    def __init__(self, address: IPv6Address):
        # This option remains constant, so create a singleton that can be re-used
        option = ServerUnicastOption(server_address=address)
        option.validate()

        super().__init__(option, always_send=True)

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, self.option.server_address)

    def pre(self, bundle: TransactionBundle):
        """
        Set flag to let the server know that unicast is ok, otherwise RejectUnwantedUnicastHandler will reject it later.

        :param bundle: The transaction bundle
        """
        bundle.allow_unicast = True


class ServerUnicastOptionHandlerFactory(HandlerFactory):
    """
    Create a ServerUnicastOptionHandler
    """

    def create(self) -> ServerUnicastOptionHandler:
        """
        Create a RequireMulticastHandler
        """
        return ServerUnicastOptionHandler(self.address)



"""
Basic handlers for options
"""

import logging

from typing import Iterable, Optional, Type

from dhcpkit.ipv6.options import OptionRequestOption, Option
from dhcpkit.ipv6.server.handlers import Handler
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


class CopyOptionHandler(Handler):
    """
    This handler just copies a type of option from the request to the response

    :param option_class: The option class to copy
    :param always_send: Always send this option, even if the OptionRequestOption doesn't ask for it
    """

    def __init__(self, option_class: Type[Option], *, always_send: bool = False):
        super().__init__()

        self.option_class = option_class
        """The class of the option from the request to the response"""

        self.always_send = always_send
        """Whether an :class:`.OptionRequestOption` in the request should be ignored"""

    def handle(self, bundle: TransactionBundle):
        """
        Copy the option from the request to the response.

        :param bundle: The transaction bundle
        """
        # Make sure this option can go into this type of response
        if not bundle.response.may_contain(self.option_class):
            return

        # Check what the client requested
        if not self.always_send:
            # Don't add if the client doesn't request it
            oro = bundle.request.get_option_of_type(OptionRequestOption)
            # noinspection PyUnresolvedReferences
            if oro and self.option_class.option_type not in oro.requested_options:
                # Client doesn't want this
                return

        # Make sure this option isn't present and then copy those from the request
        bundle.response.options = [existing_option for existing_option in bundle.response.options
                                   if not isinstance(existing_option, self.option_class)]
        bundle.response.options[:0] = bundle.request.get_options_of_type(self.option_class)


class SimpleOptionHandler(Handler):
    """
    Standard handler for simple static options

    :param option: The option instance to add to the response
    :param append: Always add, even if an option of this class already exists
    :param always_send: Always send this option, even if the OptionRequestOption doesn't ask for it
    """

    def __init__(self, option: Option, *, append: bool = False, always_send: bool = False):
        super().__init__()

        self.option = option
        """The option instance to add to the response"""

        self.option_class = type(option)
        """The class of the option"""

        self.append = append
        """Always add, even if an option of this class already exists"""

        self.always_send = always_send
        """Always send this option, even if the :class:`.OptionRequestOption` doesn't ask for it"""

    def combine(self, existing_options: Iterable[Option]) -> Optional[Option]:
        """
        If an option of this type already exists this method can combine the existing option with our own option to
        create a combined option.

        :param existing_options: The existing options
        :return: The combined option which will replace all existing options, or None to leave the existing options
        """
        # By default we just leave the existing options alone
        return None

    def handle(self, bundle: TransactionBundle):
        """
        Add the option to the response in the bundle.

        :param bundle: The transaction bundle
        """
        # Make sure this option can go into this type of response
        if not bundle.response.may_contain(self.option):
            return

        # Check what the client requested
        if not self.always_send:
            # Don't add if the client doesn't request it
            oro = bundle.request.get_option_of_type(OptionRequestOption)
            if oro and self.option.option_type not in oro.requested_options:
                # Client doesn't want this
                return

        if self.append:
            # Just add
            add = True
        else:
            # See if this option was already present
            found = bundle.response.get_options_of_type(self.option_class)
            if found:
                result = self.combine(existing_options=found)
                if isinstance(result, Option):
                    # A new option, remove the old ones
                    for option in found:
                        bundle.response.options.remove(option)

                    # And add the combined one
                    bundle.response.options.append(result)

                # Don't add the standard version
                add = False
            else:
                # The option didn't exist yet, just add it
                add = True

        if add:
            # We always want to add it, or it didn't exist yet
            bundle.response.options.append(self.option)


class OverwriteOptionHandler(Handler):
    """
    Overwriting handler for simple static options.

    :param option: The option instance to use
    :param always_send: Always send this option, even if the OptionRequestOption doesn't ask for it
    """

    def __init__(self, option: Option, *, always_send: bool = False):
        super().__init__()

        self.option = option
        """The option to add to the response"""

        self.option_class = type(option)
        """The class of the option"""

        self.always_send = always_send
        """Whether an :class:`.OptionRequestOption` in the request should be ignored"""

    def handle(self, bundle: TransactionBundle):
        """
        Overwrite the option in the response in the bundle.

        :param bundle: The transaction bundle
        """
        # Make sure this option can go into this type of response
        if not bundle.response.may_contain(self.option):
            return

        # Check what the client requested
        if not self.always_send:
            # Don't add if the client doesn't request it
            oro = bundle.request.get_option_of_type(OptionRequestOption)
            if oro and self.option.option_type not in oro.requested_options:
                # Client doesn't want this
                return

        # Make sure this option isn't present and then add our own
        bundle.response.options = [existing_option for existing_option in bundle.response.options
                                   if not isinstance(existing_option, self.option_class)]
        bundle.response.options.insert(0, self.option)



"""
Handlers for the options defined in dhcpkit.ipv6.extensions.sntp
"""

from ipaddress import IPv6Address

from typing import Iterable

from dhcpkit.ipv6.extensions.sntp import SNTPServersOption
from dhcpkit.ipv6.server.handlers.basic import SimpleOptionHandler


class SNTPServersOptionHandler(SimpleOptionHandler):
    """
    Handler for putting SNTPServersOptions in responses
    """

    def __init__(self, sntp_servers: Iterable[IPv6Address], always_send: bool = False):
        option = SNTPServersOption(sntp_servers=sntp_servers)
        option.validate()

        super().__init__(option, always_send=always_send)

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, ', '.join(map(str, self.option.sntp_servers)))

    def combine(self, existing_options: Iterable[SNTPServersOption]) -> SNTPServersOption:
        """
        Combine multiple options into one.

        :param existing_options: The existing options to include NTP servers from
        :return: The combined option
        """
        sntp_servers = []

        # Add from existing options first
        for option in existing_options:
            for sntp_server in option.sntp_servers:
                if sntp_server not in sntp_servers:
                    sntp_servers.append(sntp_server)

        # Then add our own
        for sntp_server in self.option.sntp_servers:
            if sntp_server not in sntp_servers:
                sntp_servers.append(sntp_server)

        # And return a new option with the combined addresses
        return SNTPServersOption(sntp_servers=sntp_servers)



"""
This provides a backwards-compatibility layer for the Python typing system as described in PEP484
"""

try:
    # Try importing from Python's typing module.
    # Versions before 3.5.2 however don't support Type[], so test for that one explicitly.
    from typing import Type
    from typing import *
except ImportError:
    # Fall back to the version distributed with DHCPKit, either because:
    # - the typing module didn't exist
    # - Type[] was missing
    from .py352_typing import Type
    from .py352_typing import *



"""
Test the SolMaxRTOption and InfMaxRTOption option implementations
"""
import unittest

from dhcpkit.ipv6.extensions.sol_max_rt import SolMaxRTOption, InfMaxRTOption
from tests.ipv6.options import test_option


class SolMaxRTOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('005200041a2b3c4d')
        self.option_object = SolMaxRTOption(sol_max_rt=439041101)
        self.parse_option()

    def test_validate_sol_max_rt(self):
        self.check_unsigned_integer_property('sol_max_rt', 32)

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must have length 4'):
            SolMaxRTOption.parse(bytes.fromhex('005200031a2b3c4d'))

        with self.assertRaisesRegex(ValueError, 'must have length 4'):
            SolMaxRTOption.parse(bytes.fromhex('005200051a2b3c4d00'))


class InfMaxRTOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('005300041a2b3c4d')
        self.option_object = InfMaxRTOption(inf_max_rt=439041101)
        self.parse_option()

    def test_validate_inf_max_rt(self):
        self.check_unsigned_integer_property('inf_max_rt', 32)

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must have length 4'):
            InfMaxRTOption.parse(bytes.fromhex('005300031a2b3c4d'))

        with self.assertRaisesRegex(ValueError, 'must have length 4'):
            InfMaxRTOption.parse(bytes.fromhex('005300051a2b3c4d00'))


if __name__ == '__main__':
    unittest.main()



import unittest

from dhcpkit.ipv6.messages import RelayForwardMessage, RelayReplyMessage
from dhcpkit.ipv6.server.handlers import RelayHandler
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle


class TestRelayHandler(RelayHandler):
    def handle_relay(self, bundle: TransactionBundle,
                     relay_message_in: RelayForwardMessage, relay_message_out: RelayReplyMessage):
        pass


class RelayHandlerTestCase(unittest.TestCase):
    def test_str(self):
        handler = TestRelayHandler()
        self.assertEqual(str(handler), 'TestRelayHandler')

if __name__ == '__main__':
    unittest.main()



"""
Test the RequestMessage implementation
"""
import codecs
import unittest
from ipaddress import IPv6Network, IPv6Address

from dhcpkit.ipv6.duids import LinkLayerDUID
from dhcpkit.ipv6.extensions.dns import OPTION_DNS_SERVERS
from dhcpkit.ipv6.extensions.ntp import OPTION_NTP_SERVER
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, IAPrefixOption, OPTION_IA_PD
from dhcpkit.ipv6.extensions.sntp import OPTION_SNTP_SERVERS
from dhcpkit.ipv6.extensions.sol_max_rt import OPTION_SOL_MAX_RT, OPTION_INF_MAX_RT
from dhcpkit.ipv6.messages import ConfirmMessage
from dhcpkit.ipv6.options import ElapsedTimeOption, ClientIdOption, IANAOption, \
    OptionRequestOption, OPTION_IA_NA, OPTION_VENDOR_OPTS, VendorClassOption, \
    IAAddressOption
from tests.ipv6.messages import test_client_server_message

""

confirm_message = ConfirmMessage(
    transaction_id=bytes.fromhex('f350d6'),
    options=[
        ElapsedTimeOption(elapsed_time=104),
        ClientIdOption(duid=LinkLayerDUID(hardware_type=1, link_layer_address=bytes.fromhex('3431c43cb2f1'))),
        IANAOption(iaid=bytes.fromhex('c43cb2f1'), options=[
            IAAddressOption(address=IPv6Address('2001:db8:ffff:1:c::e09c'), preferred_lifetime=375, valid_lifetime=600),
        ]),
        IAPDOption(iaid=bytes.fromhex('c43cb2f1'), options=[
            IAPrefixOption(prefix=IPv6Network('2001:db8:ffcc:fe00::/56'), preferred_lifetime=375, valid_lifetime=600),
        ]),
        OptionRequestOption(requested_options=[
            OPTION_DNS_SERVERS,
            OPTION_NTP_SERVER,
            OPTION_SNTP_SERVERS,
            OPTION_IA_PD,
            OPTION_IA_NA,
            OPTION_VENDOR_OPTS,
            OPTION_SOL_MAX_RT,
            OPTION_INF_MAX_RT,
        ]),
        VendorClassOption(enterprise_number=872),
    ],
)

confirm_packet = codecs.decode(
    '04'
    'f350d6'
    '000800020068'
    '0001000a000300013431c43cb2f1'
    '00030028c43cb2f100000000000000000005001820010db8ffff0001000c00000000e09c0000017700000258'
    '00190029c43cb2f10000000000000000001a001900000177000002583820010db8ffccfe000000000000000000'
    '0006001000170038001f00190003001100520053'
    '0010000400000368', 'hex'
)


class RequestMessageTestCase(test_client_server_message.ClientServerMessageTestCase):
    def setUp(self):
        self.packet_fixture = confirm_packet
        self.message_fixture = confirm_message
        self.parse_packet()


if __name__ == '__main__':
    unittest.main()



"""
Test the SNTP options implementations
"""
import unittest
from ipaddress import IPv6Address

from dhcpkit.ipv6.extensions.sntp import SNTPServersOption
from tests.ipv6.options import test_option


class SNTPServersOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('001f0020'
                                          '20010db8000000000000000000000001'
                                          '20010db8000000000000000000000002')
        self.option_object = SNTPServersOption(sntp_servers=[IPv6Address('2001:db8::1'),
                                                             IPv6Address('2001:db8::2')])
        self.parse_option()

    def test_validate_sntp_servers(self):
        self.option.sntp_servers = IPv6Address('2001:db8::1')
        with self.assertRaisesRegex(ValueError, 'must be a list'):
            self.option.validate()

        self.option.sntp_servers = ['2001:db8::1', '2001:db8::2']
        with self.assertRaisesRegex(ValueError, 'IPv6 address'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'length must be a multiple of 16'):
            SNTPServersOption.parse(bytes.fromhex('001f000f20010db8000000000000000000000001'))

        with self.assertRaisesRegex(ValueError, 'longer than the available buffer'):
            SNTPServersOption.parse(bytes.fromhex('001f001120010db8000000000000000000000001'))

        with self.assertRaisesRegex(ValueError, 'length must be a multiple of 16'):
            SNTPServersOption.parse(bytes.fromhex('001f001120010db800000000000000000000000100'))


if __name__ == '__main__':
    unittest.main()



"""
A socket to control the DHCPKit server
"""
import errno
import logging
import os
import socket
import time

import dhcpkit
from typing import List

logger = logging.getLogger(__name__)


class ControlConnection:
    """
    A connection of the remote control socket
    """

    def __init__(self, sock: socket.socket):
        logger.debug("Starting new control connection")

        self.sock = sock
        self.buffer = b''
        self.last_activity = time.time()

        # Set socket options
        self.sock.setblocking(False)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 100 * 1024)

        # Send welcome
        self.send("DHCPKit DHCPv6 Server {}".format(dhcpkit.__version__))

    def fileno(self) -> int:
        """
        The fileno of the listening socket, so this object can be used by select()

        :return: The file descriptor
        """
        return self.sock.fileno()

    def get_commands(self) -> List[str]:
        """
        Receive data until the next newline and return the result

        :return: A list of commands
        """
        received = None
        try:
            # Receive little bits so the remote end can't overload us with control commands
            received = self.sock.recv(64)
            self.buffer += received

            # We had activity
            self.last_activity = time.time()
        except BlockingIOError:
            # Nothing to read? Fine...
            pass

        if received == b'':
            # The other end closed the connection
            logger.debug("Control connection closed without saying goodbye")
            return [None]

        # Search for commands
        commands = []
        while True:
            parts = self.buffer.split(b'\n', maxsplit=1)
            if len(parts) < 2:
                break

            # Found one, keep the remaining buffer and return the command lowercase
            self.buffer = parts[1]
            try:
                command = parts[0].decode('utf-8').lower()
            except UnicodeError:
                # Ignore non-UTF-8 input
                continue

            if command:
                commands.append(command)

        return commands

    def send(self, output: str, eol=b'\n'):
        """
        Send data over the socket

        :param output: The data to send
        :param eol: The end-of-line character
        """
        try:
            self.sock.send(output.encode('utf-8') + eol)
        except BrokenPipeError:
            # They have gone away, fine
            pass

    def close(self):
        """
        Close the socket nicely
        """
        logger.debug("Closing control connection")
        self.sock.close()

    def acknowledge(self, feedback:str=None):
        """
        Acknowledge the command
        """
        if feedback:
            self.send('OK: {}'.format(feedback))
        else:
            self.send('OK')

    def reject(self, *args):
        """
        Reject the command
        """
        self.send('UNKNOWN')


class ControlSocket:
    """
    Remote control of the DHCPKit server
    """

    def __init__(self, socket_path: str):
        self.socket_path = socket_path

        # Create the socket and listen on it
        self.listen_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.listen_socket.setblocking(False)

        try:
            logger.info("Creating control socket {}".format(socket_path))
            self.listen_socket.bind(socket_path)
        except FileNotFoundError:
            raise RuntimeError("The path to control socket {} doesn't exist".format(socket_path)) from None
        except OSError as e:
            if e.errno == errno.EADDRINUSE:
                logger.debug("Control socket at {} exists, trying to see if it's still alive".format(socket_path))

                # Is this an old socket? Try to connect
                try:
                    test_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                    test_socket.connect(socket_path)
                except OSError as e2:
                    if e2.errno == errno.ECONNREFUSED:
                        # Nobody listening, just delete it and try again
                        logger.debug("Replacing old control socket {}".format(socket_path))
                        os.unlink(socket_path)
                        self.listen_socket.bind(socket_path)
                    elif e2.errno == errno.ENOTSOCK:
                        raise RuntimeError("Control socket {} is unusable".format(socket_path)) from None

        if not self.listen_socket.getsockname():
            raise RuntimeError("Cannot create control socket {}".format(socket_path)) from None

        self.listen_socket.listen(32)

    def close(self):
        """
        Close the socket nicely
        """
        self.listen_socket.close()

    def fileno(self) -> int:
        """
        The fileno of the listening socket, so this object can be used by select()

        :return: The file descriptor
        """
        return self.listen_socket.fileno()

    def accept(self) -> ControlConnection:
        """
        Accept a new connection

        :return: The new connection
        """
        try:
            sock = self.listen_socket.accept()[0]
            return ControlConnection(sock)
        except OSError:
            logger.debug("Control connection broken after connecting, ignoring")
            return None



"""
Configuration file definition and parsing
"""
import inspect
import logging
import os

import ZConfig.info
from ZConfig import SchemaResourceError
from ZConfig.loader import SchemaLoader, ConfigLoader
from ZConfig.schema import BaseParser

from dhcpkit.ipv6.server.config_elements import MainConfig
from dhcpkit.ipv6.server.extension_registry import server_extension_registry

logger = logging.getLogger(__name__)


def get_config_loader() -> ConfigLoader:
    """
    Get the config loader with all extensions

    :return: The fully extended config loader
    """

    # Patch the parser because otherwise it will reject the example section in the schema

    # Construct the paths to all necessary files
    schema_filename = os.path.abspath(os.path.join(os.path.dirname(__file__), "config_schema.xml"))

    # Patch the parser if we have an old version of ZConfig
    # noinspection PyProtectedMember
    if 'schema' not in BaseParser._allowed_parents['example']:
        # noinspection PyProtectedMember
        BaseParser._allowed_parents['example'] += ['schema', 'sectiontype']

        # Add some properties
        ZConfig.info.SchemaType.example = None
        ZConfig.info.SectionType.example = None

        # Patch a method by saving the old method and replacing the original with a patched version
        ZConfig.info.oldCreateDerivedSchema = ZConfig.info.createDerivedSchema

        # noinspection PyUnresolvedReferences,PyPep8Naming
        def patchedCreateDerivedSchema(base: ZConfig.info.SchemaType) -> ZConfig.info.SchemaType:
            """
            Also copy the example section.

            :param base: The original
            :return: The copy
            """
            new = ZConfig.info.oldCreateDerivedSchema(base)
            new.example = base.example
            return new

        ZConfig.info.createDerivedSchema = patchedCreateDerivedSchema

    # Load the schema
    schema_loader = SchemaLoader()
    schema = schema_loader.loadURL(url=schema_filename)

    # Build the config loader based on the schema, extended with the schemas of option handlers
    config_loader = ConfigLoader(schema=schema)

    # Iterate over all server extensions
    for extension_name, extension in server_extension_registry.items():
        # Option handlers that refer to packages contain components
        if inspect.ismodule(extension) and hasattr(extension, '__path__'):
            # It's a package! Try to import
            try:
                config_loader.importSchemaComponent(extension.__name__)
                logger.debug("Configuration extension {} loaded".format(extension_name))
            except SchemaResourceError:
                # Component missing, assume it's a package without a config component
                pass

    return config_loader


def load_config(config_filename: str) -> MainConfig:
    """
    Load the given configuration file.

    :param config_filename: The configuration file
    :return: The parsed config
    """
    logger.debug("Loading configuration file {}".format(config_filename))

    config_loader = get_config_loader()
    config_filename = os.path.realpath(config_filename)
    config, handlers = config_loader.loadURL(config_filename)

    return config



"""
The server extension registry
"""
from dhcpkit.registry import Registry


class ServerExtensionRegistry(Registry):
    """
    Registry for DHCPKit IPv6 Server Extensions
    """
    entry_point = 'dhcpkit.ipv6.server.extensions'


# Instantiate the extension registry
server_extension_registry = ServerExtensionRegistry()



"""
Utility functions for the DHCP server
"""
import logging
import netifaces

from dhcpkit.ipv6.duids import LinkLayerDUID
from dhcpkit.utils import normalise_hex

logger = logging.getLogger(__name__)


def determine_local_duid() -> LinkLayerDUID:
    """
    Calculate our own DUID based on one of our MAC addresses

    :return: The server DUID
    """
    for interface_name in netifaces.interfaces():
        link_addresses = netifaces.ifaddresses(interface_name).get(netifaces.AF_LINK, [])
        link_addresses = [link_address['addr'] for link_address in link_addresses if link_address.get('addr')]

        for link_address in link_addresses:
            try:
                # Build a DUID from this address
                ll_addr = bytes.fromhex(normalise_hex(link_address))
                if len(ll_addr) != 6:
                    # If it is not 6 bytes long then it is not an ethernet MAC address
                    continue

                # Assume it's ethernet, build a DUID
                duid = LinkLayerDUID(hardware_type=1, link_layer_address=ll_addr)

                logger.debug("Using server DUID based on {} link address: {}".format(interface_name, link_address))

                return duid
            except ValueError:
                # Try the next one
                pass

    # We didn't find a useful server DUID
    raise ValueError("Cannot find a usable server DUID")



"""
Test the camelcase conversion functions
"""
import unittest

from dhcpkit.utils import normalise_hex


class NormaliseHexTestCase(unittest.TestCase):
    def test_hex(self):
        self.assertEqual(normalise_hex(''), '')
        self.assertEqual(normalise_hex('1a2b3c'), '1a2b3c')
        self.assertEqual(normalise_hex('1a:2b:3c'), '1a2b3c')
        self.assertEqual(normalise_hex('1a:2b3c'), '1a2b3c')
        self.assertEqual(normalise_hex('1a2b:3c'), '1a2b3c')

    def test_hex_with_colons(self):
        self.assertEqual(normalise_hex('', include_colons=True), '')
        self.assertEqual(normalise_hex('1a2b3c', include_colons=True), '1a:2b:3c')
        self.assertEqual(normalise_hex('1a:2b:3c', include_colons=True), '1a:2b:3c')
        self.assertEqual(normalise_hex('1a:2b3c', include_colons=True), '1a:2b:3c')
        self.assertEqual(normalise_hex('1a2b:3c', include_colons=True), '1a:2b:3c')

    def test_bad_hex(self):
        with self.assertRaisesRegex(ValueError, 'not valid hex'):
            normalise_hex('1a2:b3c')

        with self.assertRaisesRegex(ValueError, 'not valid hex'):
            normalise_hex('Something')


if __name__ == '__main__':
    unittest.main()



"""
Configuration elements for the dns option handlers
"""
from dhcpkit.ipv6.server.extensions.sntp import SNTPServersOptionHandler
from dhcpkit.ipv6.server.handlers import HandlerFactory


class SNTPServersOptionHandlerFactory(HandlerFactory):
    """
    Create the handler for SNTP servers.
    """

    def create(self) -> SNTPServersOptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """
        return SNTPServersOptionHandler(self.addresses, always_send=self.always_send)



"""
Test the VendorSpecificInformationOption implementation
"""
import unittest

from dhcpkit.ipv6.options import VendorSpecificInformationOption
from tests.ipv6.options import test_option


class VendorSpecificInformationOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0011002000009d10') \
                            + b'\x00\x01\x00\x06Option' \
                            + b'\x00\x02\x00\x0eAnother Option'
        self.option_object = VendorSpecificInformationOption(40208, [(1, b'Option'), (2, b'Another Option')])
        self.parse_option()

    def test_enterprise_number(self):
        self.check_unsigned_integer_property('enterprise_number', size=32)

    def test_vendor_options(self):
        self.option.vendor_options = b'Not a list'
        with self.assertRaisesRegex(ValueError, 'must be a list'):
            self.option.validate()

        self.option.vendor_options = [b'In a list', b'X' * 2 ** 16]
        with self.assertRaisesRegex(ValueError, 'must be a list'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'length does not match'):
            VendorSpecificInformationOption.parse(bytes.fromhex('0011000d00009d10') + b'\x00\x01\x00\x06Option')

        with self.assertRaisesRegex(ValueError, 'length does not match'):
            VendorSpecificInformationOption.parse(bytes.fromhex('0011000f00009d10') +
                                                  b'\x00\x01\x00\x06Option\x00\x00\x00\x00')


if __name__ == '__main__':
    unittest.main()



"""
Common logging component
"""

# Extra log levels
import logging

DEBUG_HANDLING = 7
DEBUG_PACKETS = 5

logging.addLevelName(DEBUG_PACKETS, 'PACKET')
logging.addLevelName(DEBUG_HANDLING, 'HANDLING')



"""
Test the ReconfigureMessageOption implementation
"""
import unittest

from dhcpkit.ipv6.messages import MSG_RENEW
from dhcpkit.ipv6.options import ReconfigureMessageOption
from tests.ipv6.options import test_option


class ReconfigureMessageOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0013000105')
        self.option_object = ReconfigureMessageOption(message_type=MSG_RENEW)
        self.parse_option()

    def test_message_type(self):
        self.option.message_type = 0
        with self.assertRaisesRegex(ValueError, 'type must be'):
            self.option.validate()

        self.option.message_type = 4
        with self.assertRaisesRegex(ValueError, 'type must be'):
            self.option.validate()

        self.option.message_type = 5
        self.option.validate()

        self.option.message_type = 6
        with self.assertRaisesRegex(ValueError, 'type must be'):
            self.option.validate()

        self.option.message_type = 10
        with self.assertRaisesRegex(ValueError, 'type must be'):
            self.option.validate()

        self.option.message_type = 11
        self.option.validate()

        self.option.message_type = 12
        with self.assertRaisesRegex(ValueError, 'type must be'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must have length 1'):
            ReconfigureMessageOption.parse(bytes.fromhex('0013000005'))

        with self.assertRaisesRegex(ValueError, 'must have length 1'):
            ReconfigureMessageOption.parse(bytes.fromhex('001300020500'))


if __name__ == '__main__':
    unittest.main()



"""
Handlers for the options defined in dhcpkit.ipv6.extensions.ntp
"""
from typing import Iterable

from dhcpkit.ipv6.extensions.ntp import NTPSubOption, NTPServersOption
from dhcpkit.ipv6.server.handlers.basic import SimpleOptionHandler


class NTPServersOptionHandler(SimpleOptionHandler):
    """
    Handler for putting NTPServersOption in responses
    """

    def __init__(self, sub_options: Iterable[NTPSubOption], always_send: bool = False):
        option = NTPServersOption(options=sub_options)
        option.validate()

        super().__init__(option, always_send=always_send)

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, ', '.join([option.value for option in self.option.options]))

    def combine(self, existing_options: Iterable[NTPServersOption]) -> NTPServersOption:
        """
        Combine multiple options into one.

        :param existing_options: The existing options to include NTP servers from
        :return: The combined option
        """
        sub_options = []

        # Add from existing options first
        for option in existing_options:
            for sub_option in option.options:
                if sub_option not in sub_options:
                    sub_options.append(sub_option)

        # Then add our own
        for sub_option in self.option.dns_servers:
            if sub_option not in sub_options:
                sub_options.append(sub_option)

        # And return a new option with the combined addresses
        return NTPServersOption(options=sub_options)



"""
Datatypes useful for the logging component
"""
import logging
import logging.handlers
import socket

from ZConfig.datatypes import RangeCheckedConversion

from dhcpkit.common.server.logging import DEBUG_PACKETS, DEBUG_HANDLING


def syslog_facility(value: str) -> int:
    """
    Convert the strings representing syslog facilities to their numerical value

    :param value: The string representing the syslog facility
    :return: Numerical syslog facility
    """
    lower_value = value.lower()
    if lower_value not in logging.handlers.SysLogHandler.facility_names:
        raise ValueError("'{}' is not a valid syslog facility name".format(value))
    return logging.handlers.SysLogHandler.facility_names[lower_value]


def logging_level(value: str) -> int:
    """
    Convert the strings representing logging levels to their numerical value

    :param value: The string representing the logging level
    :return: Numerical logging level
    """
    name_to_level = {
        'critical': logging.CRITICAL,
        'error': logging.ERROR,
        'warn': logging.WARNING,
        'warning': logging.WARNING,
        'info': logging.INFO,
        'debug': logging.DEBUG,
        'debug-packets': DEBUG_PACKETS,
        'debug-handling': DEBUG_HANDLING,
        'notset': logging.NOTSET,
    }

    lookup_value = value.lower().replace('_', '-')
    if lookup_value not in name_to_level:
        raise ValueError("'{}' is not a valid log level".format(value))
    return name_to_level[lookup_value]


def udp_or_tcp(value: str) -> int:
    """
    Convert the strings "udp" and "tcp" to SOCK_DGRAM and SOCK_STREAM respectively

    :param value: The string "udp" or "tcp"
    :return: SOCK_DGRAM or SOCK_STREAM
    """
    lower_value = value.lower()
    if lower_value in ('udp', 'dgram'):
        return socket.SOCK_DGRAM
    elif lower_value in ('tcp', 'stream'):
        return socket.SOCK_STREAM
    else:
        raise ValueError("The protocol must be UDP (or DGRAM) or TCP (or STREAM)")


def rotation_style(value: str) -> str:
    """
    Determine the rotation style.

    :param value: String representation of rotation style
    :return: Normalised rotation style
    """
    lower_value = value.lower()
    # return 's'
    if lower_value in ('hour', 'hourly'):
        return 'h'
    elif lower_value in ('day', 'daily'):
        return 'd'
    elif lower_value in ('week', 'weekly'):
        return 'w'
    elif lower_value == 'size':
        return 'SIZE'
    else:
        raise ValueError("Rotation style must be hourly, daily, weekly or size")


rotation_count = RangeCheckedConversion(int, min=1)



"""
Filter on subnet that the link address is in
"""
from ipaddress import IPv6Network

from cached_property import cached_property

from dhcpkit.ipv6.server.filters import Filter, FilterFactory
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle
from dhcpkit.utils import camelcase_to_dash


class SubnetFilter(Filter):
    """
    Filter on subnet that the link address is in
    """

    @cached_property
    def filter_description(self) -> str:
        """
        A short description of this filter for log messages.

        :return: The description
        """
        simple_name = camelcase_to_dash(self.__class__.__name__)
        if simple_name.endswith('-filter'):
            simple_name = simple_name[:-7]

        return "{} in {}".format(simple_name, [str(prefix) for prefix in self.filter_condition])

    def match(self, bundle: TransactionBundle) -> bool:
        """
        Check if the link-address is in the subnet

        :param bundle: The transaction bundle
        :return: Whether the link-address matches
        """
        # Check if the link-address is in any of the prefixes
        return any([bundle.link_address in prefix for prefix in self.filter_condition])


class SubnetFilterFactory(FilterFactory):
    """
    Create a subnet filter
    """
    name_datatype = staticmethod(IPv6Network)
    filter_class = SubnetFilter

    @property
    def filter_condition(self):
        """
        Return the filter condition, the list of prefixes
        :return: The filter condition
        """
        return [self.name]


class SubnetGroupFilterFactory(FilterFactory):
    """
    Create a subnet filter
    """
    filter_class = SubnetFilter

    @property
    def filter_condition(self):
        """
        Return the filter condition, the list of prefixes
        :return: The filter condition
        """
        return self.prefixes



"""
Configuration section for LinkLayerTimeDUID
"""
import logging

from dhcpkit.ipv6.duids import LinkLayerTimeDUID

logger = logging.getLogger(__name__)


def duid_llt(section) -> LinkLayerTimeDUID:
    """
    Create a LinkLayerDUID from the data provided in the config section.

    :param section: The section data
    :return: The DUID object
    """
    duid = LinkLayerTimeDUID(hardware_type=section.hardware_type,
                             link_layer_address=section.link_layer_address,
                             time=section.timestamp)
    duid.validate()
    return duid



"""
Test the Message implementation
"""
import unittest

from dhcpkit.ipv6.messages import Message, UnknownMessage


class MessageTestCase(unittest.TestCase):
    def setUp(self):
        # The following attributes must be overruled by child classes
        # The basics are tested with a simple UnknownMessage
        self.packet_fixture = bytes.fromhex('ff') + b'ThisIsAnUnknownMessage'
        self.message_fixture = UnknownMessage(255, b'ThisIsAnUnknownMessage')
        self.parse_packet()

    def parse_packet(self):
        self.length, self.message = Message.parse(self.packet_fixture)
        self.assertIsInstance(self.message, Message)
        self.message_class = type(self.message)

    def test_length(self):
        self.assertEqual(self.length, len(self.packet_fixture))

    def test_parse(self):
        self.assertEqual(self.message, self.message_fixture)

    def test_save_parsed(self):
        self.assertEqual(self.packet_fixture, self.message.save())

    def test_save_fixture(self):
        self.assertEqual(self.packet_fixture, self.message_fixture.save())

    def test_validate(self):
        # This should be ok
        self.message.validate()

    def check_unsigned_integer_property(self, property_name: str, size: int = None):
        """
        Perform basic verification of validation of an unsigned integer

        :param property_name: The property under test
        :param size: The number of bits of this integer field
        """
        # Do the basic integer checks
        setattr(self.message, property_name, 0.1)
        with self.assertRaisesRegex(ValueError, 'integer'):
            self.message.validate()

        setattr(self.message, property_name, 0)
        self.message.validate()

        setattr(self.message, property_name, -1)
        with self.assertRaisesRegex(ValueError, 'unsigned .* integer'):
            self.message.validate()

        if not size:
            # We can't do any further tests without knowing the size
            return

        setattr(self.message, property_name, 2 ** size - 1)
        self.message.validate()

        setattr(self.message, property_name, 2 ** size)
        with self.assertRaisesRegex(ValueError, 'unsigned {} bit integer'.format(size)):
            self.message.validate()


if __name__ == '__main__':
    unittest.main()



"""
Test the UserClassOption implementation
"""
import unittest

from dhcpkit.ipv6.options import UserClassOption
from tests.ipv6.options import test_option


class UserClassOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('000f0016') + b'\x00\x05Class' + b'\x00\x0dAnother Class'
        self.option_object = UserClassOption([b'Class', b'Another Class'])
        self.parse_option()

    def test_user_classes(self):
        self.option.user_classes = b'Not a list'
        with self.assertRaisesRegex(ValueError, 'must be a list'):
            self.option.validate()

        self.option.user_classes = [b'In a list', b'X' * 2 ** 16]
        with self.assertRaisesRegex(ValueError, 'must be a list'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'length does not match'):
            UserClassOption.parse(bytes.fromhex('000f0006') + b'\x00\x05Class')

        with self.assertRaisesRegex(ValueError, 'length does not match'):
            UserClassOption.parse(bytes.fromhex('000f0009') + b'\x00\x05Class\x00\x01X')


if __name__ == '__main__':
    unittest.main()



"""
Some messages need a status code in the response. These handlers insert that status code if no other handler did.
"""
from dhcpkit.ipv6.messages import ConfirmMessage, ReleaseMessage, DeclineMessage
from dhcpkit.ipv6.options import StatusCodeOption, STATUS_SUCCESS
from dhcpkit.ipv6.server.handlers import Handler
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle


class AddMissingStatusOptionHandler(Handler):
    """
    The handler that makes sure that replies to confirm messages have a status code. When we reach the end without any
    status code being set we assume success. Other option handlers set the status to something else if they cannot
    confirm their part.
    """

    def handle(self, bundle: TransactionBundle):
        """
        Update the status of the reply to :class:`.ConfirmMessage`, :class:`.ReleaseMessage` and
        :class:`.DeclineMessage`.

        :param bundle: The transaction bundle
        """
        if isinstance(bundle.request, ConfirmMessage):
            message = "Assignments confirmed"
        elif isinstance(bundle.request, ReleaseMessage):
            message = "Thank you for releasing your resources"
        elif isinstance(bundle.request, DeclineMessage):
            message = "Our apologies for assigning you unusable addresses"
        else:
            # Not a message type we're interested in
            return

        existing = bundle.response.get_option_of_type(StatusCodeOption)
        if not existing:
            bundle.response.options.append(
                StatusCodeOption(STATUS_SUCCESS, status_message=message)
            )



"""
An option handler that assigns addresses based on DUID from a CSV file
"""
import codecs
import csv
import logging
from ipaddress import IPv6Address, IPv6Network

from typing import Mapping, List, Tuple

from dhcpkit.ipv6.duids import DUID
from dhcpkit.ipv6.extensions.remote_id import RemoteIdOption
from dhcpkit.ipv6.options import ClientIdOption, InterfaceIdOption
from dhcpkit.ipv6.server.extensions.static_assignments import StaticAssignmentHandler, Assignment
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle
from dhcpkit.utils import normalise_hex

logger = logging.getLogger(__name__)


class CSVStaticAssignmentHandler(StaticAssignmentHandler):
    """
    Assign addresses and/or prefixes based on the contents of a CSV file
    """

    def __init__(self, filename: str,
                 address_preferred_lifetime: int, address_valid_lifetime: int,
                 prefix_preferred_lifetime: int, prefix_valid_lifetime: int):
        """
        Initialise the mapping. This handler will respond to clients on responsible_for_links and assume that all
        addresses in the mapping are appropriate for on those links.

        :param filename: The filename containing the CSV data
        """
        super().__init__(address_preferred_lifetime, address_valid_lifetime,
                         prefix_preferred_lifetime, prefix_valid_lifetime)

        self.filename = filename
        self.mapping = self.read_csv_file(filename)

    def __str__(self):
        return "{} from {}".format(self.__class__.__name__, self.filename)

    def get_assignment(self, bundle: TransactionBundle) -> Assignment:
        """
        Look up the assignment based on DUID, Interface-ID of the relay closest to the client and Remote-ID of the
        relay closest to the client, in that order.

        :param bundle: The transaction bundle
        :return: The assignment, if any
        """
        # Look up based on DUID
        duid_option = bundle.request.get_option_of_type(ClientIdOption)
        duid = 'duid:' + codecs.encode(duid_option.duid.save(), 'hex').decode('ascii')
        if duid in self.mapping:
            return self.mapping[duid]

        # Look up based on Interface-ID
        interface_id_option = bundle.incoming_relay_messages[0].get_option_of_type(InterfaceIdOption)
        if interface_id_option:
            interface_id = 'interface-id:' + codecs.encode(interface_id_option.interface_id, 'hex').decode('ascii')
            if interface_id in self.mapping:
                return self.mapping[interface_id]

        # Look up based on Remote-ID
        remote_id_option = bundle.incoming_relay_messages[0].get_option_of_type(RemoteIdOption)
        if remote_id_option:
            remote_id = 'remote-id:{}:{}'.format(remote_id_option.enterprise_number,
                                                 codecs.encode(remote_id_option.remote_id, 'hex').decode('ascii'))
            if remote_id in self.mapping:
                return self.mapping[remote_id]

        # Nothing found
        return Assignment(address=None, prefix=None)

    def read_csv_file(self, csv_filename: str) -> Mapping[str, Assignment]:
        """
        Read the assignments from the file specified in the configuration

        :param csv_filename: The filename of the CSV file
        :return: A dictionary mapping identifiers to assignments
        """
        assignments = dict(self.parse_csv_file(csv_filename))
        logger.info("Loaded {} assignments from {}".format(len(assignments), csv_filename))
        return assignments

    @staticmethod
    def parse_csv_file(csv_filename: str) -> List[Tuple[str, Assignment]]:
        """
        Read the assignments from the file specified in the configuration

        :param csv_filename: The filename of the CSV file
        :return: An list of identifiers and their assignment
        """

        logger.debug("Loading assignments from {}".format(csv_filename))

        with open(csv_filename) as csv_file:
            # Auto-detect the CSV dialect
            sniffer = csv.Sniffer()
            sample = csv_file.read(10240)
            dialect = sniffer.sniff(sample)

            # Restart and parse
            csv_file.seek(0)
            reader = csv.DictReader(csv_file, dialect=dialect)

            # First line is column headings
            for row in reader:
                try:
                    address_str = row['address'].strip()
                    address = address_str and IPv6Address(address_str) or None

                    prefix_str = row['prefix'].strip()
                    prefix = prefix_str and IPv6Network(prefix_str) or None

                    # Validate and normalise id input
                    row_id = row['id']

                    if row_id.startswith('duid:'):
                        duid_hex = row_id.split(':', 1)[1]
                        duid_bytes = codecs.decode(duid_hex, 'hex')
                        length, duid = DUID.parse(duid_bytes, length=len(duid_bytes))
                        duid_hex = codecs.encode(duid.save(), 'hex').decode('ascii')
                        row_id = 'duid:{}'.format(duid_hex)

                    elif row_id.startswith('interface-id:'):
                        interface_id_hex = row_id.split(':', 1)[1]
                        interface_id_hex = normalise_hex(interface_id_hex)
                        interface_id = codecs.decode(interface_id_hex, 'hex')
                        interface_id_hex = codecs.encode(interface_id, 'hex').decode('ascii')
                        row_id = 'interface_id:{}'.format(interface_id_hex)

                    elif row_id.startswith('interface-id-str:'):
                        interface_id = row_id.split(':', 1)[1]
                        interface_id_hex = codecs.encode(interface_id.encode('ascii'), 'hex').decode('ascii')
                        row_id = 'interface_id:{}'.format(interface_id_hex)

                    elif row_id.startswith('remote-id:') or row_id.startswith('remote-id-str:'):
                        remote_id_data = row_id.split(':', 1)[1]
                        try:
                            enterprise_id, remote_id = remote_id_data.split(':', 1)
                            enterprise_id = int(enterprise_id)
                            if row_id.startswith('remote-id:'):
                                remote_id = normalise_hex(remote_id)
                                remote_id = codecs.decode(remote_id, 'hex')
                            else:
                                remote_id = remote_id.encode('ascii')

                            row_id = 'remote-id:{}:{}'.format(enterprise_id,
                                                              codecs.encode(remote_id, 'hex').decode('ascii'))
                        except ValueError:
                            raise ValueError("Remote-ID must be formatted as 'remote-id:<enterprise>:<remote-id-hex>', "
                                             "for example: 'remote-id:9:0123456789abcdef")

                    else:
                        raise ValueError("The id must start with duid: or interface-id: followed by a hex-encoded "
                                         "value, interface-id-str: followed by an ascii string, remote-id: followed by "
                                         "an enterprise-id, a colon and a hex-encoded value or remote-id-str: followed"
                                         "by an enterprise-id, a colon and an ascii string")

                    # Store the normalised id
                    logger.debug("Loaded assignment for {}".format(row_id))
                    yield row_id, Assignment(address=address, prefix=prefix)

                except KeyError:
                    raise ValueError("Assignment CSV must have columns 'id', 'address' and 'prefix'")
                except ValueError as e:
                    logger.error("Ignoring {} line {} with invalid value: {}".format(csv_file, reader.line_num, e))



"""
The DUID registry
"""
from dhcpkit.registry import Registry


class DUIDRegistry(Registry):
    """
    Registry for DHCPKit IPv6 DUIDs
    """
    entry_point = 'dhcpkit.ipv6.duids'

    def get_name(self, item: object) -> str:
        """
        Get the name for the by_name mapping.

        :param item: The item to determine the name of
        :return: The name to use as key in the mapping
        """
        name = super().get_name(item)

        # Remove suffixes
        if name.endswith('-duid'):
            name = name[:-5]

        return name


# Instantiate the DUID registry
duid_registry = DUIDRegistry()



"""
Filter on elapsed time indicated by the client
"""
import operator
from collections import namedtuple

from cached_property import cached_property
from dhcpkit.ipv6.options import ElapsedTimeOption
from dhcpkit.ipv6.server.filters import Filter, FilterFactory
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle
from dhcpkit.utils import camelcase_to_dash
from typing import List

TimeLimit = namedtuple('TimeLimit', ['operator', 'limit'])


class ElapsedTimeFilter(Filter):
    """
    Filter on marks that have been placed on the incoming message
    """

    @cached_property
    def filter_description(self) -> str:
        """
        A short description of this filter for log messages.

        :return: The description
        """
        simple_name = camelcase_to_dash(self.__class__.__name__)
        if simple_name.endswith('-filter'):
            simple_name = simple_name[:-7]

        return simple_name + ' ' + ' and '.join(["{operator} {value}".format(operator=condition.operator.__name__,
                                                                             value=condition.limit)
                                                 for condition in self.filter_condition])

    def match(self, bundle: TransactionBundle) -> bool:
        """
        Check if the elapsed time is within the configured limits

        :param bundle: The transaction bundle
        :return: Whether the elapsed time is within the limits
        """
        elapsed_time_option = bundle.request.get_option_of_type(ElapsedTimeOption)
        if not elapsed_time_option:
            # If there is no elapsed time then ignore the request
            return False

        for time_limit in self.filter_condition:
            if not time_limit.operator(elapsed_time_option.elapsed_time, time_limit.limit):
                # It failed the check
                return False

        # Apparently they all match
        return True


class ElapsedTimeFilterFactory(FilterFactory):
    """
    Create a MarkedWithFilter
    """
    filter_class = ElapsedTimeFilter

    def validate_config_section(self):
        """
        Check that at least one filter condition is provided, and that if multiple conditions are provided they are
        compatible with each other.
        """
        if not self.less_than and not self.more_than:
            raise ValueError("At least one time limit ('more-than' or 'less-than') must be configured")

        if self.less_than and self.more_than and self.less_than <= self.more_than:
            raise ValueError("The timing limits of 'more-than' and 'less-than' conflict")

    @property
    def filter_condition(self) -> List[TimeLimit]:
        """
        The filter condition is based on the configured time limits.

        :return: A list of time limits
        """
        limits = []
        if self.more_than:
            # ElapsedTime contains the time in 1/100 of a second, but we just use second precision, so multiply
            limits.append(TimeLimit(operator=operator.gt, limit=self.more_than * 100))
        if self.less_than:
            # ElapsedTime contains the time in 1/100 of a second, but we just use second precision, so multiply
            limits.append(TimeLimit(operator=operator.lt, limit=self.less_than * 100))

        return limits



"""
Filter on marks that have been placed on the incoming message
"""
from dhcpkit.ipv6.server.filters import Filter, FilterFactory
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle


class MarkedWithFilter(Filter):
    """
    Filter on marks that have been placed on the incoming message
    """

    def match(self, bundle: TransactionBundle) -> bool:
        """
        Check if the configured mark is in the set

        :param bundle: The transaction bundle
        :return: Whether the configured mark is present
        """
        return self.filter_condition in bundle.marks


class MarkedWithFilterFactory(FilterFactory):
    """
    Create a MarkedWithFilter
    """
    name_datatype = staticmethod(str)
    filter_class = MarkedWithFilter



"""
A script to generate .rst documentation based on the config schema
"""
import argparse
import io
import logging
import os
import sys
from textwrap import dedent, indent
from xml.dom import Node

from ZConfig.info import SchemaType, SectionType, AbstractType, SectionInfo, KeyInfo, MultiKeyInfo
from typing import List, Union, Iterable

from dhcpkit.ipv6.server.config_parser import get_config_loader

logger = logging.getLogger()
logging.basicConfig(level=logging.INFO)


def normalise_link_name(link: str) -> str:
    """
    Convert i.e. "filter_factory" to "filters"

    :param link: The original link name
    :return: The normalised link name
    """
    if link.endswith('_factory'):
        link = link[:-8] + 's'

    return link


def link_to(text: str, link: str = None) -> str:
    """
    Make the text a reference link.

    :param text: The text to link
    :param link: The link destination, if different from the text
    :return: The texts as a reference link
    """
    if link:
        link = normalise_link_name(link)
        return ':ref:`{} <{}>`'.format(text, link)
    else:
        text = normalise_link_name(text)
        return ':ref:`{}`'.format(text)


def link_destination(name: str) -> str:
    """
    Create an rst link.

    :param name: The destination to link to
    :return: The reStructuredText link
    """
    name = normalise_link_name(name)
    return '.. _{}:'.format(name)


def nicer_type_name(name: str) -> str:
    """
    Make a nicer name for a type.

    :param name: The ugly name
    :return: The nicer name
    """
    if name.endswith('_factory'):
        name = name[:-8] + 's'

    name = name.capitalize()

    return name


def heading(text: str, underline: str) -> str:
    """
    Create a heading using the specified underline character.

    :param text: The text to use as the heading
    :param underline: The character to underline with
    :return: The heading in rst format
    """
    return text + '\n' + (underline * len(text))


def create_file(name, args):
    """
    Create a file, or a file-like dummy if dry-run is enabled

    :param name: The relative file/path name
    :param args: The command like arguments
    :return: A file-like object
    """
    full_name = os.path.join(args.output_dir, name)

    if args.dry_run:
        logger.info("Dry-run, would have written to {}".format(full_name))
        return io.StringIO()

    if os.path.exists(full_name) and not args.force:
        logger.info("Skipping existing file, would have written to {}".format(full_name))
        return None

    # Make the directory, just to be sure
    os.makedirs(os.path.dirname(full_name), exist_ok=True)

    # Create and return the file
    logger.info("Creating {}".format(full_name))
    return open(full_name, 'w')


def write_lines(file, lines: Iterable[str]):
    """
    Write a set of lines to the file

    :param file: The file, or None
    :param lines: The lines to write
    """
    if file is None:
        return

    lines_with_nl = [line + '\n' for line in lines]
    file.writelines(lines_with_nl)


def reindent(text: str, new_indent: str = '') -> str:
    """
    Fix the indentation.

    :param text: The original text with unknown indentation
    :param new_indent: The string to indent with
    :return: The text with fixed indentation
    """
    if text is None:
        return ''

    # Split lines
    lines = text.split('\n')

    # If the first line is not indented then don't include it in the dedent
    if not lines[0].startswith((' ', '\t')):
        output = lines.pop(0)
    else:
        output = ''

    # Join the rest together
    text = '\n'.join(lines)

    # Dedent (remove common indents)
    text = dedent(text)

    if text:
        output += '\n' + text

    return indent(output, new_indent)


def key_doc(info: Union[KeyInfo, MultiKeyInfo, SectionInfo]) -> List[str]:
    """
    Generate documentation for a key.

    :param info: The information object for this key
    :return: The documentation for that key
    """

    if info.name == '+':
        title = '<multiple>'
    else:
        title = str(info.name)

    # Determine extra flags
    extras = []
    if info.minOccurs > 0:
        extras += ['required']
    if info.maxOccurs > 1:
        extras += ['multiple allowed']
    if isinstance(info, SectionInfo):
        extras += ['section of type {}'.format(link_to(info.sectiontype.name))]

    if extras:
        title += ' ({})'.format(', '.join(extras))

    output = [
        title,
        reindent(info.description, '    '),
        ''
    ]

    if info.example:
        if '\n' in info.example:
            # Multi-line example
            output += [
                '    **Example**:',
                '',
                '    .. code-block:: dhcpkitconf',
                '',
                reindent(info.example, '        '),
                '',
            ]
        else:
            # Single-line example
            output += [
                '    **Example**: "{}"'.format(reindent(str(info.example))),
                '',
            ]

    default = info.getdefault()
    if default:
        # If it is a list with only one element then pretend it's not a list
        if isinstance(default, list) and len(default) == 1:
            default = default[0]

        # Format the default value(s)
        if isinstance(default, list):
            output += (
                ['    **Default**:',
                 ''] +
                ['    - "{}"'.format(reindent(str(item.value))) for item in default] +
                ['']
            )
        else:
            output += [
                '    **Default**: "{}"'.format(reindent(str(default.value))),
                '',
            ]
    elif info.metadefault:
        output += [
            '    **Default**: {}'.format(reindent(str(info.metadefault))),
            '',
        ]

    return output


def sectiontype_doc(section: SectionType) -> List[str]:
    """
    Extract the documentation for the given section.

    :param section: The section to extract documentation from
    :return: A list of strings with documentation
    """

    output = []

    # noinspection PyUnresolvedReferences
    if section.example:
        # noinspection PyUnresolvedReferences
        output += ['',
                   heading('Example', '-'),
                   '',
                   '.. code-block:: dhcpkitconf',
                   '',
                   reindent(section.example, '    '),
                   '']

    section_parameters = [(key, info) for key, info in section if key and isinstance(info, (KeyInfo, MultiKeyInfo,
                                                                                            SectionInfo))]
    subsection_types = [(key, info) for key, info in section if key is None and isinstance(info, SectionInfo)]

    if section_parameters:
        if section.name:
            output += [link_destination(section.name + '_parameters'),
                       '',
                       heading('Section parameters', '-'),
                       '']
        else:
            output += [link_destination('schema_parameters'),
                       '',
                       heading('Configuration options', '-'),
                       '']

        for key, info in section_parameters:
            output += key_doc(info)

    if subsection_types:
        output += [heading('Possible sub-section types', '-'),
                   '']

        for key, info in subsection_types:
            output += [link_to(nicer_type_name(info.sectiontype.name), info.sectiontype.name),
                       reindent(info.sectiontype.description, '    '),
                       '']

    return output


def handle_args(args: Iterable[str]):
    """
    Handle the command line arguments.

    :param args: Command line arguments
    :return: The arguments object
    """
    parser = argparse.ArgumentParser(
        description="Generate DHCPKit configuration documentation for Sphinx.",
    )

    parser.add_argument("-o", "--output-dir", metavar="DESTDIR", required=True, help="Directory to place all output")
    parser.add_argument("-e", "--extension", metavar="PACKAGE", default='dhcpkit', help="Document the given extension")

    group = parser.add_mutually_exclusive_group()
    group.add_argument("-f", "--force", action="store_true", help="Overwrite existing files")
    group.add_argument("-n", "--dry-run", action="store_true", help="Run the script without creating files")

    args = parser.parse_args(args)

    return args


def main(args: Iterable[str]) -> int:
    """
    Generate .rst documentation based on the config schema

    :param args: Command line arguments
    :return: Program exit code
    """
    # Handle the command line arguments
    args = handle_args(args)

    # Get the configuration schema with extensions
    config_loader = get_config_loader()
    schema = config_loader.schema
    assert isinstance(schema, SchemaType)

    # The index
    index_file = create_file('index.rst', args)
    if args.extension != 'dhcpkit':
        # ZConfig doesn't save the component description, so we'll have to find it ourselves
        description = ''

        # noinspection PyProtectedMember
        components = schema._components.items()
        for name, url in components:
            # Check the package name
            if not name.startswith('package:{}.'.format(args.extension)):
                continue

            # Open the component XML
            component = config_loader.openResource(url)
            from xml.dom.minidom import parse
            dom = parse(component.file)

            # Find the description node
            for node in dom.documentElement.childNodes:
                if node.nodeType != Node.ELEMENT_NODE or node.tagName != 'description':
                    continue

                # We found the description! Separate new description with a blank line if there is already some text
                if description:
                    description += '\n\n'

                for child_node in node.childNodes:
                    if child_node.nodeType == Node.TEXT_NODE:
                        description += child_node.nodeValue

                # Clean up the whitespace
                description = description.strip()
                break

        if not description:
            description = 'This is the documentation of the configuration options of the {} package.'.format(
                args.extension)

        # Only include this in the main configuration, not in extension docs
        write_lines(index_file, [
            heading('IPv6 Server extension configuration', '='),
            '',
            reindent(description),
            '',
        ])
    else:
        # Only include this in the main configuration, not in extension docs
        write_lines(index_file, [
            heading('IPv6 Server configuration', '='),
            '',
            reindent(schema.description),
            '',
            '.. toctree::',
            '',
            '    config_file',
            '',
        ])

        # Put the main config format in a separate file
        config_file = create_file('config_file.rst', args)
        write_lines(config_file, [
            heading('Configuration file format', '='),
            '',
            reindent(schema.description),
            '',
        ] + sectiontype_doc(schema))

    # Keep track of which section types are covered under abstract types
    handled_section_types = set()

    # Collect types
    abstract_types = set()
    section_types = set()
    for type_name in schema.gettypenames():
        section_type = schema.gettype(type_name)

        if isinstance(section_type, AbstractType):
            # Remember which section types are covered here
            handled_section_types.update(set(section_type.getsubtypenames()))
            abstract_types.add(section_type)
        elif isinstance(section_type, SectionType):
            # If we are limited to a single extension then try to determine if this section type belongs to that
            # extension by looking at the datatype. It isn't pretty but it works...
            datatype = getattr(section_type, 'datatype', None)
            if not datatype.__module__.startswith(args.extension + '.'):
                continue

            section_types.add(section_type)

    # Only include section types with no underscore in the name
    section_types = [section_type for section_type in section_types if '_' not in section_type.name]
    section_types = sorted(section_types, key=lambda t: t.name)

    # Only include section types we have collected before
    documented_names = set([section_type.name for section_type in section_types])
    abstract_types = [abstract_type for abstract_type in abstract_types
                      if documented_names.intersection(abstract_type.getsubtypenames())]
    abstract_types = sorted(abstract_types, key=lambda t: t.name)

    # These are the section types that are not under an abstract section type
    root_types = [section_type for section_type in section_types if section_type.name not in handled_section_types]

    if root_types:
        # Table of contents
        write_lines(index_file, [
            heading('Overview of sections', '-'),
            '',
            '.. toctree::',
            '    :maxdepth: 1',
            '',
        ])

        # A file for each section type that hasn't been handled already
        for section_type in root_types:
            # Write a reference to the index
            write_lines(index_file, ['    ' + section_type.name])

            # Write the file
            file = create_file(section_type.name + '.rst', args)
            write_lines(file, [
                link_destination(section_type.name),
                '',
                heading(nicer_type_name(section_type.name), '='),
                '',
                reindent(section_type.description),
                ''
            ])
            write_lines(file, sectiontype_doc(section_type))

    if abstract_types:
        # Table of contents
        write_lines(index_file, [
            '',
            heading('Overview of section types', '-'),
            '',
            '.. toctree::',
            '    :maxdepth: 2',
            '',
        ])

        # A file for each abstract type
        for section_type in abstract_types:
            subtypes = section_type.getsubtypenames()

            # Only include section types we have collected before
            subtypes = [subtype for subtype in subtypes if subtype in documented_names]

            if not subtypes:
                # Don't generate documentation for abstract types with no implementation
                continue

            # Write a reference to the index
            write_lines(index_file, ['    ' + section_type.name])

            # Write the file
            file = create_file(section_type.name + '.rst', args)
            write_lines(file, [
                link_destination(section_type.name),
                '',
                heading(nicer_type_name(section_type.name), '='),
                '',
                reindent(section_type.description),
                '',
                '.. toctree::',
                '',
            ])

            # Write the implementations
            for subtype_name in subtypes:
                subtype = section_type.getsubtype(subtype_name)

                write_lines(file, ['    ' + subtype.name])

                sub_file = create_file(subtype.name + '.rst', args)
                write_lines(sub_file, [
                    link_destination(subtype.name),
                    '',
                    heading(nicer_type_name(subtype.name), '='),
                    '',
                    reindent(subtype.description),
                    ''
                ] + sectiontype_doc(subtype))


def run() -> int:
    """
    Run the main program and handle exceptions

    :return: The program exit code
    """
    try:
        # Run the server
        return main(sys.argv[1:])
    except Exception as e:
        logger.critical("Error: {}".format(e))
        return 1


if __name__ == '__main__':
    sys.exit(run())



"""
Config processing for a handler to echo a RemoteIdOption back to the relay
"""
from dhcpkit.ipv6.server.extensions.remote_id import CopyRemoteIdOptionHandler
from dhcpkit.ipv6.server.handlers import HandlerFactory


class CopyRemoteIdOptionHandlerFactory(HandlerFactory):
    """
    Config processing for a handler to echo a RemoteIdOption back to the relay
    """
    def create(self) -> CopyRemoteIdOptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """
        return CopyRemoteIdOptionHandler()



"""
Factory for the implementation of a listener on a unicast address of a local network interface
"""
import logging
import netifaces
import socket
from ipaddress import IPv6Address

from ZConfig.matcher import SectionValue
from typing import Iterable

from dhcpkit.ipv6 import SERVER_PORT
from dhcpkit.ipv6.server.listeners import Listener, ListenerFactory
from dhcpkit.ipv6.utils import is_global_unicast

logger = logging.getLogger(__name__)


class UnicastListenerFactory(ListenerFactory):
    """
    Factory for the implementation of a listener on a unicast address of a local network interface
    """

    name_datatype = staticmethod(IPv6Address)

    def __init__(self, section: SectionValue):
        # Auto-detect the interface name that the specified address is on
        self.found_interface = None

        super().__init__(section)

    def validate_config_section(self):
        """
        Validate the interface information
        """
        # Validate what the user supplied
        if not is_global_unicast(self.name):
            raise ValueError("The listener address must be a global unicast address")

        for interface_name in netifaces.interfaces():
            interface_addresses = [IPv6Address(addr_info['addr'].split('%')[0])
                                   for addr_info
                                   in netifaces.ifaddresses(interface_name).get(netifaces.AF_INET6, [])]

            if self.name in interface_addresses:
                self.found_interface = interface_name
                break

        if not self.found_interface:
            raise ValueError("Cannot find address {} on any interface".format(self.name))

    def create(self, old_listeners: Iterable[Listener] = None) -> Listener:
        """
        Create a listener of this class based on the configuration in the config section.

        :param old_listeners: A list of existing listeners in case we can recycle them
        :return: A listener object
        """
        # Try recycling
        old_listeners = list(old_listeners or [])
        for old_listener in old_listeners:
            if self.match_socket(sock=old_listener.listen_socket, address=self.name):
                logger.debug("Recycling existing socket for {} on {}".format(self.name, self.found_interface))
                sock = old_listener.listen_socket
                break
        else:
            logger.debug("Creating socket for {} on {}".format(self.name, self.found_interface))
            sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            sock.bind((str(self.name), SERVER_PORT))

        return Listener(self.found_interface, sock, marks=self.marks)



"""
The basic configuration objects
"""

import grp
import logging

from dhcpkit.common.server.config_elements import ConfigSection
from dhcpkit.ipv6.server.message_handler import MessageHandler
from dhcpkit.ipv6.server.utils import determine_local_duid

logger = logging.getLogger(__name__)


class MainConfig(ConfigSection):
    """
    The top level configuration element
    """

    def clean_config_section(self):
        """
        Clean up the config, making sure we have user, group and DUID
        """
        if self.section.group is None:
            # No group specified
            try:
                self.section.group = grp.getgrgid(self.section.user.pw_gid)
            except KeyError:
                raise ValueError("User {} has a non-existent primary group {}".format(self.section.user.pw_name,
                                                                                      self.section.user.pw_gid))

        if not self.section.server_id:
            self.section.server_id = determine_local_duid()

    def create_message_handler(self) -> MessageHandler:
        """
        Create a message handler based on this configuration.

        :return: The message handler
        """
        sub_filters = []
        for filter_factory in self.section.filter_factories:
            sub_filters.append(filter_factory())

        sub_handlers = []
        for handler_factory in self.section.handler_factories:
            sub_handlers.append(handler_factory())

        return MessageHandler(self.section.server_id, sub_filters, sub_handlers, self.section.allow_rapid_commit)


class StatisticsConfig(ConfigSection):
    """
    Configuration of the statistics gatherer
    """



"""
Code to keep the receiving and sending sockets together. When receiving traffic on a link-local multicast address
the reply should be sent from a link-local address on the receiving interface. This class makes it easy to keep those
together.
"""
import abc
import logging
import socket
from ipaddress import IPv6Address

from dhcpkit.common.server.config_elements import ConfigElementFactory
from dhcpkit.common.server.logging import DEBUG_PACKETS
from dhcpkit.ipv6 import SERVER_PORT, CLIENT_PORT
from typing import Iterable

logger = logging.getLogger(__name__)

# A global counter for log message correlation
message_counter = 0


class ListeningSocketError(Exception):
    """
    Signal that the listening socket could not be created.
    """


# Create optimised classes to store packets and metadata in
class IncomingPacketBundle:
    """
    A class that is very efficient to pickle because this is what will be sent to worker processes.

    Using a class instead of a namedtuple makes it easier to extend it in the future. To make this possible all
    properties should have a default value, and the constructor must be called with keyword arguments only.
    """

    def __init__(self, *, message_id: str = '????', data: bytes = b'', sender: IPv6Address = None,
                 link_address: IPv6Address = None, interface_name: str = '', received_over_multicast: bool = False,
                 marks: Iterable[str] = None):
        """
        Store the provided data

        :param message_id: An identifier for logging to correlate log-messages
        :param data: The bytes received from the listener
        :param sender: The IPv6 address of the sender
        :param link_address: The IPv6 address to identify the link that the packet was received over
        :param interface_name: The interface-ID  to identify the link that the packet was received over
        :param received_over_multicast: Whether this packet was received over multicast
        :param marks: A list of marks, usually set by the listener based on the configuration
        """
        self.message_id = message_id
        self.data = data
        self.sender = sender
        self.link_address = link_address or IPv6Address(0)
        self.interface_name = interface_name
        self.received_over_multicast = received_over_multicast,
        self.marks = list(marks or [])

    @property
    def interface_id(self) -> bytes:
        """
        Backwards compatibility bytes representation of the interface name

        :return: The interface name in UTF-8
        """
        return self.interface_name.encode('utf-8')

    def __getstate__(self):
        return (self.message_id, self.data, self.sender, self.link_address, self.interface_name,
                self.received_over_multicast, self.marks)

    def __setstate__(self, state):
        (self.message_id, self.data, self.sender, self.link_address, self.interface_name,
         self.received_over_multicast, self.marks) = state


class OutgoingPacketBundle:
    """
    A class that is very efficient to pickle because this is what will be received back from worker processes.

    Using a class instead of a namedtuple makes it easier to extend it in the future. To make this possible all
    properties should have a default value, and the constructor must be called with keyword arguments only.
    """

    def __init__(self, *, message_id: str = '????', data: bytes = b'', destination: IPv6Address = None,
                 port: int = CLIENT_PORT):
        """
        Store the provided data

        :param message_id: An identifier for logging to correlate log-messages
        :param data: The bytes that should be sent by the listener
        :param destination: The IPv6 address of the destination
        :param port: The port number the packet should be sent to
        """
        self.message_id = message_id
        self.data = data
        self.destination = destination
        self.port = port

    def __getstate__(self):
        """
        Pickle the state as a tuple.
        """
        return self.message_id, self.data, self.destination, self.port

    def __setstate__(self, state):
        """
        Parse the state as a tuple.
        """
        self.message_id, self.data, self.destination, self.port = state


class Listener:
    """
    A wrapper for a normal socket that bundles a socket to listen on with a (potentially different) socket
    to send replies from.

    :param interface_name: The name of the interface
    :param listen_socket: The socket we are listening on, may be a unicast or multicast socket
    :param reply_socket: The socket replies are sent from, must be a unicast socket
    :param global_address: The global address on the listening interface

    :type interface_name: str
    :type interface_index: int
    :type listen_socket: socket.socket
    :type listen_address: IPv6Address
    :type reply_socket: socket.socket
    :type reply_address: IPv6Address
    :type global_address: IPv6Address
    """

    def __init__(self, interface_name: str, listen_socket: socket.socket, reply_socket: socket.socket = None,
                 global_address: IPv6Address = None, marks: Iterable[str] = None):
        self.interface_name = interface_name
        self.interface_id = interface_name.encode('utf-8')
        self.listen_socket = listen_socket
        self.reply_socket = reply_socket
        self.marks = list(marks or [])
        if self.reply_socket is None:
            self.reply_socket = self.listen_socket

        # Check that we have IPv6 UDP sockets
        if self.listen_socket.family != socket.AF_INET6 or self.listen_socket.proto != socket.IPPROTO_UDP \
                or self.reply_socket.family != socket.AF_INET6 or self.reply_socket.proto != socket.IPPROTO_UDP:
            raise ListeningSocketError("Listen and reply sockets have to be IPv6 UDP sockets")

        listen_sockname = self.listen_socket.getsockname()
        reply_sockname = self.reply_socket.getsockname()

        # Check that we are on the right port
        if listen_sockname[1] != SERVER_PORT or reply_sockname[1] != SERVER_PORT:
            raise ListeningSocketError("Listen and reply sockets have to be on port {}".format(SERVER_PORT))

        # Check that they are both on the same interface
        if listen_sockname[3] != reply_sockname[3]:
            raise ListeningSocketError("Listen and reply sockets have to be on same interface")

        self.interface_index = listen_sockname[3]
        self.listen_address = IPv6Address(listen_sockname[0].split('%')[0])
        self.reply_address = IPv6Address(reply_sockname[0].split('%')[0])

        if global_address:
            self.global_address = global_address
        elif not self.listen_address.is_link_local and not self.listen_address.is_multicast:
            self.global_address = self.listen_address
        else:
            raise ListeningSocketError("Cannot determine global address on interface {}".format(self.interface_name))

        # We only support fixed address binding
        if self.listen_address.is_unspecified or self.reply_address.is_unspecified:
            raise ListeningSocketError("This server only supports listening on explicit address, not on wildcard")

        # Multicast listeners must have link-local reply addresses
        if self.listen_address.is_multicast and not self.reply_address.is_link_local:
            raise ListeningSocketError("Multicast listening addresses need link-local reply socket")

        # Non-multicast listeners need to use a single address
        if not self.listen_address.is_multicast and self.reply_socket != self.listen_socket:
            raise ListeningSocketError("Unicast listening addresses can't use separate reply sockets")

    def recv_request(self) -> IncomingPacketBundle:
        """
        Receive incoming messages

        :return: The address of the sender of the message and the received message
        """
        data, sender = self.listen_socket.recvfrom(65536)

        # Update the message counter and wrap it if necessary
        global message_counter
        message_counter += 1
        if message_counter > 0xFFFFFF:
            message_counter = 1

        # Create the message-ID
        message_id = '#{:06X}'.format(message_counter)

        logger.log(DEBUG_PACKETS, "{message_id}: Received message from {client_addr} port {port} on {interface}".format(
            message_id=message_id,
            client_addr=sender[0],
            port=sender[1],
            interface=self.interface_name))

        return IncomingPacketBundle(message_id=message_id,
                                    data=data,
                                    sender=IPv6Address(sender[0].split('%')[0]),
                                    link_address=self.global_address,
                                    interface_name=self.interface_name,
                                    received_over_multicast=self.listen_address.is_multicast,
                                    marks=self.marks)

    def send_reply(self, packet: OutgoingPacketBundle) -> bool:
        """
        Send a reply using the information in the outer RelayReplyMessage

        :param packet: The packet to reply with
        :return: Whether sending has succeeded
        """

        destination = (str(packet.destination), packet.port, 0, self.interface_index)

        sent_length = self.reply_socket.sendto(packet.data, destination)
        success = len(packet.data) == sent_length

        if success:
            logger.log(DEBUG_PACKETS, "{message_id}: Sent message to {client_addr} port {port} on {interface}".format(
                message_id=packet.message_id,
                client_addr=packet.destination,
                port=packet.port,
                interface=self.interface_name))
        else:
            logger.error("{message_id}: Could not send message to {client_addr} port {port} on {interface}".format(
                message_id=packet.message_id,
                client_addr=packet.destination,
                port=packet.port,
                interface=self.interface_name))

        return success

    def fileno(self) -> int:
        """
        The fileno of the listening socket, so this object can be used by select()

        :return: The file descriptor
        """
        return self.listen_socket.fileno()


class ListenerFactory(ConfigElementFactory, metaclass=abc.ABCMeta):
    """
    Base class for listener factories
    """

    @staticmethod
    def match_socket(sock: socket.socket, address: IPv6Address, interface: int = 0) -> bool:
        """
        Determine if we can recycle this socket

        :param sock: An existing socket
        :param address: The address we want
        :param interface: The interface number we want
        :return: Whether the socket is suitable
        """
        if sock.family != socket.AF_INET6 or sock.type != socket.SOCK_DGRAM or sock.proto != socket.IPPROTO_UDP:
            # Different protocol
            return False

        sockname = sock.getsockname()
        if IPv6Address(sockname[0].split('%')[0]) != address or sockname[1] != SERVER_PORT or sockname[3] != interface:
            # Wrong address
            return False

        # Amazing! This one seems to match
        return True



"""
Configuration section for LinkLayerDUID
"""
import logging

from dhcpkit.ipv6.duids import LinkLayerDUID

logger = logging.getLogger(__name__)


def duid_ll(section) -> LinkLayerDUID:
    """
    Create a LinkLayerDUID from the data provided in the config section.

    :param section: The section data
    :return: The DUID object
    """
    duid = LinkLayerDUID(hardware_type=section.hardware_type,
                         link_layer_address=section.link_layer_address)
    duid.validate()
    return duid



"""
Test the ReconfigureAcceptOption implementation
"""
import unittest

from dhcpkit.ipv6.options import ReconfigureAcceptOption
from tests.ipv6.options import test_option


class ReconfigureAcceptOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('00140000')
        self.option_object = ReconfigureAcceptOption()
        self.parse_option()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must have length 0'):
            ReconfigureAcceptOption.parse(bytes.fromhex('00140001'))


if __name__ == '__main__':
    unittest.main()



"""
Option handlers that inserts a PreferenceOption in replies
"""
from dhcpkit.ipv6.options import PreferenceOption
from dhcpkit.ipv6.server.handlers import HandlerFactory
from dhcpkit.ipv6.server.handlers.basic import SimpleOptionHandler


class PreferenceOptionHandler(SimpleOptionHandler):
    """
    The handler for PreferenceOption which adds a preference option to appropriate responses
    """

    def __init__(self, preference: int):
        # This option remains constant, so create a singleton that can be re-used
        option = PreferenceOption(preference=preference)
        option.validate()

        super().__init__(option, always_send=True)

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, self.option.preference)


class PreferenceOptionHandlerFactory(HandlerFactory):
    """
    Create an IgnoreRequestHandler
    """

    def create(self) -> PreferenceOptionHandler:
        """
        Create an IgnoreRequestHandler
        """
        return PreferenceOptionHandler(preference=self.level)



"""
Server extension to handle prefix delegation options properly
"""
import logging

from typing import List

from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, STATUS_NOPREFIXAVAIL, IAPrefixOption
from dhcpkit.ipv6.messages import SolicitMessage, RequestMessage, RenewMessage, ReleaseMessage, RebindMessage
from dhcpkit.ipv6.options import StatusCodeOption, STATUS_NOBINDING
from dhcpkit.ipv6.server.handlers import Handler, CannotRespondError
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


def create_cleanup_handlers() -> List[Handler]:
    """
    Create handlers to clean up stuff in the transaction bundle

    :return: Handlers to add to the handler chain
    """
    return [UnansweredIAPDOptionHandler()]


class UnansweredIAPDOptionHandler(Handler):
    """
    A handler that answers to all unanswered IAPDOptions

    :param authoritative: Whether this handler is authorised to tell clients to stop using prefixes
    """

    def __init__(self, authoritative: bool = True):
        super().__init__()
        self.authoritative = authoritative

    def handle(self, bundle: TransactionBundle):
        """
        Make sure that every :class:`.IAPDOption` is answered.

        :param bundle: The transaction bundle
        """
        for option in bundle.get_unhandled_options(IAPDOption):
            if isinstance(bundle.request, (SolicitMessage, RequestMessage)):
                # If the delegating router will not assign any prefixes to any IA_PDs in a subsequent Request from the
                # requesting router, the delegating router MUST send an Advertise message to the requesting router that
                # includes the IA_PD with no prefixes in the IA_PD and a Status Code option in the IA_PD containing
                # status code NoPrefixAvail and a status message for the user
                #
                # We do the same for unanswered requests
                bundle.response.options.append(IAPDOption(option.iaid, options=[
                    StatusCodeOption(STATUS_NOPREFIXAVAIL, "No prefixes available")
                ]))

            elif isinstance(bundle.request, RenewMessage):
                # Renew message: If the delegating router cannot find a binding for the requesting router's IA_PD the
                # delegating router returns the IA_PD containing no prefixes with a Status Code option set to
                # NoBinding in the Reply message.

                prefixes = ', '.join(map(str, option.get_prefixes()))
                logger.warning("No handler renewed {}: sending NoBinding status".format(prefixes))

                bundle.response.options.append(IAPDOption(option.iaid, options=[
                    StatusCodeOption(STATUS_NOBINDING, "No prefixes assigned to you")
                ]))

            elif isinstance(bundle.request, RebindMessage):
                # Rebind message: If the delegating router cannot find a binding for the requesting router's IA_PD and
                # the delegating router determines that the prefixes in the IA_PD are not appropriate for the link to
                # which the requesting router's interface is attached according to the delegating routers explicit
                # configuration, the delegating router MAY send a Reply message to the requesting router containing
                # the IA_PD with the lifetimes of the prefixes in the IA_PD set to zero.  This Reply constitutes an
                # explicit notification to the requesting router that the prefixes in the IA_PD are no longer valid.
                #
                # If the delegating router is unable to determine if the prefix is not appropriate for the link, the
                # Rebind message is discarded.
                #
                # The authoritative flag indicates whether this option may claim whether it is able to determine if a
                # prefix is appropriate for the link.
                if not self.authoritative:
                    raise CannotRespondError

                prefixes = ', '.join(map(str, option.get_prefixes()))
                logger.warning("No handler answered rebind of {}: withdrawing prefixes".format(prefixes))

                reply_suboptions = []
                for suboption in option.get_options_of_type(IAPrefixOption):
                    reply_suboptions.append(IAPrefixOption(suboption.prefix, preferred_lifetime=0, valid_lifetime=0))

                bundle.response.options.append(IAPDOption(option.iaid, options=reply_suboptions))

            elif isinstance(bundle.request, ReleaseMessage):
                # For each IA in the Release message for which the server has no binding information, the server adds an
                # IA option using the IAID from the Release message, and includes a Status Code option with the value
                # NoBinding in the IA option.  No other options are included in the IA option.
                bundle.response.options.append(IAPDOption(option.iaid, options=[
                    StatusCodeOption(STATUS_NOBINDING, "No prefixes assigned to you")
                ]))



"""
Test the ElapsedTimeOption implementation
"""
import unittest

from dhcpkit.ipv6.options import AuthenticationOption
from tests.ipv6.options import test_option


class AuthenticationOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('000b000d1234567890abcdef13579ae89f')
        self.option_object = AuthenticationOption(protocol=18, algorithm=52, rdm=86,
                                                  replay_detection=b'\x78\x90\xab\xcd\xef\x13\x57\x9a',
                                                  auth_info=b'\xe8\x9f')
        self.parse_option()

    def test_validate_protocol(self):
        self.check_unsigned_integer_property('protocol', size=8)

    def test_validate_algorithm(self):
        self.check_unsigned_integer_property('algorithm', size=8)

    def test_validate_rdm(self):
        self.check_unsigned_integer_property('rdm', size=8)

    def test_replay_detection(self):
        self.option.replay_detection = b'ABCDEFG'
        with self.assertRaisesRegex(ValueError, 'must contain 8 bytes'):
            self.option.validate()

        self.option.replay_detection = b'ABCDEFGH'
        self.option.validate()

        self.option.replay_detection = b'ABCDEFGHI'
        with self.assertRaisesRegex(ValueError, 'must contain 8 bytes'):
            self.option.validate()

        self.option.replay_detection = 'ABCDEFGH'
        with self.assertRaisesRegex(ValueError, 'must contain 8 bytes'):
            self.option.validate()

    def test_auth_info(self):
        self.option.auth_info = 'ABCDEFGH'
        with self.assertRaisesRegex(ValueError, 'must contain bytes'):
            self.option.validate()

        self.option.auth_info = b'ABCDEFGH'
        self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'shorter than the minimum length'):
            AuthenticationOption.parse(bytes.fromhex('000b00000000ffff'))


if __name__ == '__main__':
    unittest.main()



"""
Configuration elements for the SOL_MAX_RT option handlers
"""
from dhcpkit.ipv6.server.extensions.sol_max_rt import SolMaxRTOptionHandler, InfMaxRTOptionHandler
from dhcpkit.ipv6.server.handlers import HandlerFactory


def max_rt(value: str) -> int:
    """
    Convert the name of the section to the number of seconds, and validate the range

    :param value: Config section name
    :return: Number of seconds
    """
    seconds = int(value)

    if not (60 <= seconds <= 86400):
        raise ValueError("MAX_RT must be between 60 and 86400 seconds")

    return seconds


class SolMaxRTOptionHandlerFactory(HandlerFactory):
    """
    Create the handler for the SolMaxRTOption.
    """

    def create(self) -> SolMaxRTOptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """
        return SolMaxRTOptionHandler(self.limit, always_send=self.always_send)


class InfMaxRTOptionHandlerFactory(HandlerFactory):
    """
    Create the handler for the InfMaxRTOption.
    """

    def create(self) -> InfMaxRTOptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """
        return InfMaxRTOptionHandler(self.limit, always_send=self.always_send)



"""
Test the RelayMessageOption implementation
"""
import unittest

from dhcpkit.ipv6.message_registry import message_registry
from dhcpkit.ipv6.messages import UnknownMessage, ClientServerMessage
from dhcpkit.ipv6.options import RelayMessageOption
from tests.ipv6.options import test_option


# A dummy option that may not be in a RelayMessageOption
class NonRelayableMessage(UnknownMessage):
    pass


# Add the constraint to disallow putting it in a RelayMessageOption
RelayMessageOption.add_may_contain(NonRelayableMessage, max_occurrence=0)


# A dummy message that has the wrong length
class WeirdLengthMessage(ClientServerMessage):
    message_type = 254

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None):
        return 4


class RelayMessageOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('00090017ff') + b'ThisIsAnUnknownMessage'
        self.option_object = RelayMessageOption(relayed_message=UnknownMessage(255, b'ThisIsAnUnknownMessage'))
        self.parse_option()

    def test_validate_relayed_message(self):
        self.option.relayed_message = None
        with self.assertRaisesRegex(ValueError, 'IPv6 DHCP message'):
            self.option.validate()

        self.option.relayed_message = NonRelayableMessage(255, b'ThisIsAnUnknownMessage')
        with self.assertRaisesRegex(ValueError, 'cannot contain'):
            self.option.validate()

    def test_bad_message_length(self):
        # Add a fake message type to the registry
        message_registry[254] = WeirdLengthMessage

        with self.assertRaisesRegex(ValueError, 'different length'):
            RelayMessageOption.parse(bytes.fromhex('00090006fe1234567890'))

        # And clean it up again
        del message_registry[254]


if __name__ == '__main__':
    unittest.main()



"""
Configuration section for EnterpriseDUID
"""
import logging

from dhcpkit.ipv6.duids import EnterpriseDUID

logger = logging.getLogger(__name__)


def duid_en(section) -> EnterpriseDUID:
    """
    Create a EnterpriseDUID from the data provided in the config section.

    :param section: The section data
    :return: The DUID object
    """
    duid = EnterpriseDUID(enterprise_number=section.enterprise_number,
                          identifier=section.identifier)
    duid.validate()
    return duid



"""
Configuration elements for the IANA/IAPD timing limits.
"""
from typing import Optional

from ZConfig.datatypes import SuffixMultiplier, RangeCheckedConversion

from dhcpkit.ipv6 import INFINITY
from dhcpkit.ipv6.server.extensions.timing_limits import IANATimingLimitsHandler, IAPDTimingLimitsHandler
from dhcpkit.ipv6.server.handlers import HandlerFactory

time_interval_converter = SuffixMultiplier({
    's': 1,
    'm': 60,
    'h': 60 * 60,
    'd': 60 * 60 * 24,
    'w': 60 * 60 * 24 * 7,
})

time_range_converter = RangeCheckedConversion(conversion=time_interval_converter, min=0, max=INFINITY)


def time_value(value: str) -> int:
    """
    Cast the string INFINITY to the infinity value, otherwise convert to an integer

    :param value: The string to parse
    :return: The integer value
    """
    if value.upper() == 'INFINITY':
        return INFINITY
    return time_range_converter(value)


factor_range_converter = RangeCheckedConversion(conversion=float, min=0.0, max=1.0)


def factor_value(value: str) -> Optional[float]:
    """
    Cast the string NONE to the None value, otherwise convert to a float

    :param value: The string to parse
    :return: The float value or None
    """
    if value.upper() == 'NONE':
        return None
    return factor_range_converter(value)


class IANATimingLimitsHandlerFactory(HandlerFactory):
    """
    Create the IANATimingLimitsHandler.
    """

    def validate_config_section(self):
        """
        Check if all the values are valid in combination
        """
        # Do some basic checks for impossible values
        if self.min_t1 > self.max_t2:
            raise ValueError("min_t1 must be smaller than max_t2")

        if self.factor_t1 and self.factor_t2 and self.factor_t1 > self.factor_t2:
            raise ValueError("factor_t1 must be smaller than factor_t2")

    def create(self) -> IANATimingLimitsHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """
        return IANATimingLimitsHandler(min_t1=self.min_t1, max_t1=self.max_t1, factor_t1=self.factor_t1,
                                       min_t2=self.min_t2, max_t2=self.max_t2, factor_t2=self.factor_t2)


class IAPDTimingLimitsHandlerFactory(HandlerFactory):
    """
    Create the IAPDTimingLimitsHandler.
    """

    def validate_config_section(self):
        """
        Check if all the values are valid in combination
        """
        # Do some basic checks for impossible values
        if self.min_t1 > self.max_t2:
            raise ValueError("min_t1 must be smaller than max_t2")

        if self.factor_t1 and self.factor_t2 and self.factor_t1 > self.factor_t2:
            raise ValueError("factor_t1 must be smaller than factor_t2")

    def create(self) -> IAPDTimingLimitsHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """
        return IAPDTimingLimitsHandler(min_t1=self.min_t1, max_t1=self.max_t1, factor_t1=self.factor_t1,
                                       min_t2=self.min_t2, max_t2=self.max_t2, factor_t2=self.factor_t2)



"""
Extra datatypes for the IPv6 DHCP server
"""
import grp
import pwd

from dhcpkit.utils import encode_domain

__all__ = ['domain_name', 'user_name', 'group_name']


def domain_name(value: str) -> str:
    """
    Validate and clean a domain name.

    :param value: Domain name
    :return: Validated and cleaned domain name
    """
    # Lowercase
    value = value.lower()

    # Test validity by trying to encode it
    encode_domain(value)

    # Ok for now
    return value


def user_name(value: str) -> pwd.struct_passwd:
    """
    Validate the given user name

    :param value: User name
    :return: Resolved user
    """
    try:
        return pwd.getpwnam(value)
    except KeyError:
        raise ValueError("User with name '{}' not found".format(value))


def group_name(value: str) -> grp.struct_group:
    """
    Validate the given group name

    :param value: Group name
    :return: Resolved group
    """
    try:
        return grp.getgrnam(value)
    except KeyError:
        raise ValueError("Group with name '{}' not found".format(value))



"""
A simple handler that tells the server to ignore the request.
"""
import logging

from typing import Type, Iterable

from dhcpkit.ipv6.messages import Message
from dhcpkit.ipv6.server.handlers import Handler, CannotRespondError, HandlerFactory
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


class IgnoreRequestHandler(Handler):
    """
    A simple handler that tells the server to stop processing the request and ignore it
    """

    def __init__(self, message_types: Iterable[Type[Message]] = None):
        super().__init__()
        self.message_types = tuple(set(message_types or []))

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, ', '.join([message_type.__class__.__name__
                                                                       for message_type in self.message_types]))

    def pre(self, bundle: TransactionBundle):
        """
        Stop processing

        :param bundle: The transaction bundle
        """
        # Ignore when no type specified, or when request matches a specified type
        if not self.message_types or isinstance(bundle.request, self.message_types):
            logging.info("Configured to ignore {}".format(bundle))
            raise CannotRespondError("Ignoring request")


class IgnoreRequestHandlerFactory(HandlerFactory):
    """
    Create an IgnoreRequestHandler
    """

    def create(self) -> Handler:
        """
        Create an IgnoreRequestHandler
        """
        return IgnoreRequestHandler(self.message_types)



"""
Configuration elements for the SIP server option handlers
"""
from dhcpkit.ipv6.server.extensions.sip_servers import SIPServersDomainNameListOptionHandler, \
    SIPServersAddressListOptionHandler
from dhcpkit.ipv6.server.handlers import HandlerFactory


class SIPServersDomainNameListOptionHandlerFactory(HandlerFactory):
    """
    Create the handler for SIP servers.
    """

    def create(self) -> SIPServersDomainNameListOptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """
        return SIPServersDomainNameListOptionHandler(domain_names=self.domain_names, always_send=self.always_send)


class SIPServersAddressListOptionHandlerFactory(HandlerFactory):
    """
    Create the handler for SIP servers.
    """

    def create(self) -> SIPServersAddressListOptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """
        return SIPServersAddressListOptionHandler(sip_servers=self.addresses, always_send=self.always_send)



"""
Configuration elements for the static assignment handlers
"""
import logging

from ZConfig.datatypes import existing_file

from dhcpkit.ipv6.server.extensions.static_assignments.csv import CSVStaticAssignmentHandler
from dhcpkit.ipv6.server.extensions.static_assignments.sqlite import SqliteStaticAssignmentHandler
from dhcpkit.ipv6.server.handlers import HandlerFactory

logger = logging.getLogger(__name__)


class CSVStaticAssignmentHandlerFactory(HandlerFactory):
    """
    Factory for a handler that reads assignments from a CSV file
    """

    name_datatype = staticmethod(existing_file)

    def create(self) -> CSVStaticAssignmentHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """

        # Get the lifetimes
        address_preferred_lifetime = self.address_preferred_lifetime
        address_valid_lifetime = self.address_valid_lifetime
        prefix_preferred_lifetime = self.prefix_preferred_lifetime
        prefix_valid_lifetime = self.prefix_valid_lifetime

        return CSVStaticAssignmentHandler(
            self.name,
            address_preferred_lifetime, address_valid_lifetime,
            prefix_preferred_lifetime, prefix_valid_lifetime
        )


class SqliteStaticAssignmentHandlerFactory(HandlerFactory):
    """
    Factory for a handler that reads assignments from a SQLite database
    """

    name_datatype = staticmethod(existing_file)

    def create(self) -> SqliteStaticAssignmentHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """

        # Get the lifetimes
        address_preferred_lifetime = self.address_preferred_lifetime
        address_valid_lifetime = self.address_valid_lifetime
        prefix_preferred_lifetime = self.prefix_preferred_lifetime
        prefix_valid_lifetime = self.prefix_valid_lifetime

        return SqliteStaticAssignmentHandler(
            self.name,
            address_preferred_lifetime, address_valid_lifetime,
            prefix_preferred_lifetime, prefix_valid_lifetime
        )



"""
The NTP suboption registry
"""
from dhcpkit.registry import Registry


class NTPSuboptionRegistry(Registry):
    """
    Registry for NTP Suboptions
    """
    entry_point = 'dhcpkit.ipv6.options.ntp.suboptions'

    def get_name(self, item: object) -> str:
        """
        Get the name for the by_name mapping.

        :param item: The item to determine the name of
        :return: The name to use as key in the mapping
        """
        name = super().get_name(item)

        # Remove prefixes and suffixes
        if name.startswith('ntp-'):
            name = name[4:]
        if name.endswith('-sub-option'):
            name = name[:-11]

        return name


# Instantiate the suboption registry
ntp_suboption_registry = NTPSuboptionRegistry()



"""
Utility functions for IPv6 DHCP
"""
import logging
from ipaddress import IPv6Address, IPv6Network

from typing import Iterable, Tuple, List

from dhcpkit.ipv6.messages import Message, ClientServerMessage, RelayForwardMessage, UnknownMessage

logger = logging.getLogger(__name__)


def split_relay_chain(message: Message) -> Tuple[ClientServerMessage, List[RelayForwardMessage]]:
    """
    Separate the relay chain from the actual request message.

    :param message: The incoming message
    :returns: The request and the chain of relay messages starting with the one closest to the client
    """
    relay_messages = []
    while isinstance(message, RelayForwardMessage):
        relay_messages.insert(0, message)
        message = message.relayed_message

    # Check if we could actually read the message
    if isinstance(message, UnknownMessage):
        logger.warning("Received an unrecognised message of type {}".format(message.message_type))
        return None, None

    # Check that this message is a client->server message
    if not isinstance(message, ClientServerMessage) or not message.from_client_to_server:
        logger.warning("A server should not receive {} from a client".format(message.__class__.__name__))
        return None, None

    # Save it as the request
    return message, relay_messages


def address_in_prefixes(address: IPv6Address, prefixes: Iterable[IPv6Network]) -> bool:
    """
    Check whether the given address is part of one of the given prefixes

    :param address: The IPv6 address to check
    :param prefixes: The list of IPv6 prefixes
    :type prefixes: list[IPv6Network]
    :return: Whether the address is part of one of the prefixes
    """
    for prefix in prefixes:
        if address in prefix:
            return True

    return False


def prefix_overlaps_prefixes(prefix: IPv6Network, prefixes: Iterable[IPv6Network]) -> bool:
    """
    Check whether the given address is part of one of the given prefixes

    :param prefix: The IPv6 prefix to check
    :param prefixes: The list of IPv6 prefixes
    :type prefixes: list[IPv6Network]
    :return: Whether the address is part of one of the prefixes
    """
    for other_prefix in prefixes:
        if prefix.overlaps(other_prefix):
            return True

    return False


def is_global_unicast(address: IPv6Address) -> bool:
    """
    Check if an address is a global unicast address according to :rfc:`4291`.

    :param address: The address to check
    :return: Whether it is a global unicast address
    """
    return not (address == IPv6Address('::') or
                address == IPv6Address('::1') or
                address in IPv6Network('ff00::/8') or
                address in IPv6Network('fe80::/10'))



"""
Basic console logging based on verbosity
"""
import logging
import logging.handlers

from ZConfig.matcher import SectionValue

from dhcpkit.common.server.logging import DEBUG_HANDLING, DEBUG_PACKETS
from dhcpkit.common.server.logging.config_datatypes import logging_level


def set_verbosity_logger(logger: logging.Logger, verbosity: int, existing_console: logging.Handler = None):
    """
    Install a console based logger based based on the given verbosity.

    :param logger: The logger to add the handlers to
    :param verbosity: The verbosity level given as command line argument
    :param existing_console: The existing console handler
    """
    # Don't filter on level in the base logger
    logger.setLevel(logging.NOTSET)

    # If verbosity is 0 then leave it as it is
    if verbosity == 0:
        return

    if existing_console:
        console = existing_console
    else:
        # No console configured but verbosity asked: add a console handler
        from dhcpkit.common.server.logging.config_elements import ConsoleHandlerFactory
        fake_section = SectionValue(name='',
                                    values={'level': logging_level('warning'), 'color': None},
                                    matcher=None)
        console_factory = ConsoleHandlerFactory(fake_section)
        console = console_factory()
        logger.addHandler(console)

    # Set level according to verbosity
    if verbosity >= 5 and console.level > DEBUG_PACKETS:
        console.setLevel(DEBUG_PACKETS)
    if verbosity >= 4 and console.level > DEBUG_HANDLING:
        console.setLevel(DEBUG_HANDLING)
    if verbosity >= 3 and console.level > logging.DEBUG:
        console.setLevel(logging.DEBUG)
    elif verbosity == 2 and console.level > logging.INFO:
        console.setLevel(logging.INFO)
    elif verbosity >= 1 and console.level > logging.WARNING:
        console.setLevel(logging.WARNING)



"""
Test the ElapsedTimeOption implementation
"""
import unittest

from dhcpkit.ipv6.options import ElapsedTimeOption
from tests.ipv6.options import test_option


class ElapsedTimeOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0008000204d2')
        self.option_object = ElapsedTimeOption(elapsed_time=1234)
        self.parse_option()

    def test_validate_elapsed_time(self):
        self.check_unsigned_integer_property('elapsed_time', size=16)

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must have length 2'):
            ElapsedTimeOption.parse(bytes.fromhex('000800000000ffff'))

        with self.assertRaisesRegex(ValueError, 'must have length 2'):
            ElapsedTimeOption.parse(bytes.fromhex('000800030000ffff'))


if __name__ == '__main__':
    unittest.main()



"""
The main server process
"""
import argparse
import atexit
import fcntl
import grp
import json
import logging.handlers
import multiprocessing
import multiprocessing.queues
import os
import pwd
import selectors
import signal
import sys
import time
import types
from multiprocessing import forkserver
from multiprocessing.util import get_logger
from urllib.parse import urlparse

import dhcpkit
from ZConfig import ConfigurationSyntaxError, DataConversionError
from dhcpkit.common.server.logging.config_elements import set_verbosity_logger
from dhcpkit.ipv6.server import config_parser, queue_logger
from dhcpkit.ipv6.server.config_elements import MainConfig
from dhcpkit.ipv6.server.control_socket import ControlSocket, ControlConnection
from dhcpkit.ipv6.server.listeners import Listener, OutgoingPacketBundle
from dhcpkit.ipv6.server.statistics import ServerStatistics
from dhcpkit.ipv6.server.worker import setup_worker, handle_message
from typing import Tuple, Iterable, Optional

logger = logging.getLogger()

logging_thread = None


@atexit.register
def stop_logging_thread():
    """
    Stop the logging thread from the global
    """
    global logging_thread
    if logging_thread:
        logging_thread.stop()


def handle_args(args: Iterable[str]):
    """
    Handle the command line arguments.

    :param args: Command line arguments
    :return: The arguments object
    """
    parser = argparse.ArgumentParser(
        description="A flexible IPv6 DHCP server written in Python.",
    )

    parser.add_argument("config", help="the configuration file")
    parser.add_argument("-v", "--verbosity", action="count", default=0, help="increase output verbosity")
    parser.add_argument("-p", "--pidfile", action="store", help="save the server's PID to this file")

    args = parser.parse_args(args)

    return args


def drop_privileges(user: pwd.struct_passwd, group: grp.struct_group, permanent: bool = True):
    """
    Drop root privileges and change to something more safe.

    :param user: The tuple with user info
    :param group: The tuple with group info
    :param permanent: Whether we want to drop just the euid (temporary), or all uids (permanent)
    """
    # Restore euid=0 if we have previously changed it
    if os.geteuid() != 0 and os.getuid() == 0:
        restore_privileges()

    if os.geteuid() != 0:
        raise RuntimeError("Not running as root: cannot change uid/gid to {}/{}".format(user.pw_name, group.gr_name))

    # Remove group privileges
    os.setgroups([])

    if permanent:
        os.setgid(group.gr_gid)
        os.setuid(user.pw_uid)
    else:
        os.setegid(group.gr_gid)
        os.seteuid(user.pw_uid)

    # Ensure a very conservative umask
    os.umask(0o077)

    if permanent:
        logger.debug("Permanently dropped privileges to {}/{}".format(user.pw_name, group.gr_name))
    else:
        logger.debug("Dropped privileges to {}/{}".format(user.pw_name, group.gr_name))


def restore_privileges():
    """
    Restore root privileges
    """
    if os.getuid() != 0:
        raise RuntimeError("Privileges have been permanently dropped, cannot restore them")

    os.seteuid(0)
    os.setegid(0)

    logger.debug("Restored root privileges")


def create_pidfile(args, config: MainConfig) -> Optional[str]:
    """
    Create a PID file when configured to do so.

    :param args: The command line arguments
    :param config: The server configuration
    :return: The name of the created PID file
    """
    # Create the PID file while we are root
    if args.pidfile:
        pid_filename = os.path.realpath(args.pidfile)
    elif config.pid_file:
        pid_filename = os.path.realpath(config.pid_file)
    else:
        pid_filename = None

    if pid_filename:
        # A different umask for here
        old_umask = os.umask(0o022)
        with open(pid_filename, 'w') as pidfile:
            logger.info("Writing PID-file {}".format(pid_filename))
            pidfile.write("{}\n".format(os.getpid()))
        os.umask(old_umask)

    return pid_filename


def create_control_socket(config: MainConfig) -> ControlSocket:
    """
    Create a control socket when configured to do so.

    :param config: The server configuration
    :return: The name of the created control socket
    """
    if config.control_socket:
        socket_filename = os.path.realpath(config.control_socket)
        uid = config.control_socket_user.pw_uid
        gid = config.control_socket_group.gr_gid if config.control_socket_group else config.control_socket_user.pw_gid

        # A different umask for here
        old_umask = os.umask(0o117)
        control_socket = ControlSocket(socket_filename)
        os.chown(socket_filename, uid, gid)
        os.umask(old_umask)
        return control_socket


def create_handler_callbacks(listening_socket: Listener, message_id: str) -> Tuple[types.FunctionType,
                                                                                   types.FunctionType]:
    """
    Create a callback for the handler method that still knows the listening socket and the sender

    :param listening_socket: The listening socket to remember
    :param message_id: An identifier for logging to correlate log-messages
    :return: A callback function with the listening socket and sender enclosed
    """

    def callback(reply):
        """
        A callback that handles the result of a handler

        :param reply: The result from the handler
        """
        try:
            if reply is None:
                # No reply: we're done with this request
                return

            if not isinstance(reply, OutgoingPacketBundle):
                logger.error("{}: Handler returned invalid result, not sending a reply".format(message_id))
                return

            try:
                listening_socket.send_reply(reply)
            except ValueError as e:
                logger.error("{}: Handler returned invalid message: {}".format(message_id, e))
                return

        except Exception as e:
            # Catch-all exception handler
            logger.exception("{}: Caught unexpected exception {!r}".format(message_id, e))

    def error_callback(e: Exception):
        """
        Log an error about this exception

        :param e: The exception from the worker
        """
        logger.error("{}: Caught unexpected exception {!r}".format(message_id, e))

    return callback, error_callback


def main(args: Iterable[str]) -> int:
    """
    The main program loop

    :param args: Command line arguments
    :return: The program exit code
    """
    # Handle command line arguments
    args = handle_args(args)
    set_verbosity_logger(logger, args.verbosity)

    # Go to the working directory
    config_file = os.path.realpath(args.config)
    os.chdir(os.path.dirname(config_file))

    try:
        # Read the configuration
        config = config_parser.load_config(config_file)
    except (ConfigurationSyntaxError, DataConversionError) as e:
        # Make the config exceptions a bit more readable
        msg = e.message
        if e.lineno and e.lineno != -1:
            msg += ' on line {}'.format(e.lineno)
        if e.url:
            parts = urlparse(e.url)
            msg += ' in {}'.format(parts.path)
        logger.critical(msg)
        return 1
    except ValueError as e:
        logger.critical(e)
        return 1

    # Immediately drop privileges in a non-permanent way so we create logs with the correct owner
    drop_privileges(config.user, config.group, permanent=False)

    # Trigger the forkserver at this point, with dropped privileges, and ignoring KeyboardInterrupt
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    multiprocessing.set_start_method('forkserver')
    forkserver.ensure_running()

    # Initialise the logger
    config.logging.configure(logger, verbosity=args.verbosity)
    logger.info("Starting Python DHCPv6 server v{}".format(dhcpkit.__version__))

    # Create our selector
    sel = selectors.DefaultSelector()

    # Convert signals to messages on a pipe
    signal_r, signal_w = os.pipe()
    flags = fcntl.fcntl(signal_w, fcntl.F_GETFL, 0)
    flags = flags | os.O_NONBLOCK
    fcntl.fcntl(signal_w, fcntl.F_SETFL, flags)
    signal.set_wakeup_fd(signal_w)
    sel.register(signal_r, selectors.EVENT_READ)

    # Ignore normal signal handling by attaching dummy handlers (SIG_IGN will not put messages on the pipe)
    signal.signal(signal.SIGINT, lambda signum, frame: None)
    signal.signal(signal.SIGTERM, lambda signum, frame: None)
    signal.signal(signal.SIGHUP, lambda signum, frame: None)
    signal.signal(signal.SIGINFO, lambda signum, frame: None)

    # Excessive exception catcher
    exception_history = []

    # Some stats
    message_count = 0

    # Create a queue for our children to log to
    logging_queue = multiprocessing.Queue()

    # This will be where we store the new config after a reload
    statistics = ServerStatistics()
    listeners = []
    control_socket = None
    stopping = False
    while not stopping:
        # Safety first: assume we want to quit when we break the inner loop unless told otherwise
        stopping = True

        # Initialise the logger again
        lowest_log_level = config.logging.configure(logger, verbosity=args.verbosity)

        # Enable multiprocessing logging, mostly useful for development
        mp_logger = get_logger()
        mp_logger.propagate = config.logging.log_multiprocessing

        global logging_thread
        if logging_thread:
            logging_thread.stop()

        logging_thread = queue_logger.QueueLevelListener(logging_queue, *logger.handlers)
        logging_thread.start()

        # Restore our privileges while we write the PID file and open network listeners
        restore_privileges()

        # Open the network listeners
        old_listeners = listeners
        listeners = []
        for listener_factory in config.listener_factories:
            # Create new listener while trying to re-use existing sockets
            listeners.append(listener_factory(old_listeners + listeners))

        # Write the PID file
        pid_filename = create_pidfile(args=args, config=config)

        # Create a control socket
        if control_socket:
            sel.unregister(control_socket)
            control_socket.close()

        control_socket = create_control_socket(config=config)
        if control_socket:
            sel.register(control_socket, selectors.EVENT_READ)

        # And Drop privileges again
        drop_privileges(config.user, config.group, permanent=False)

        # Remove any file descriptors from the previous config
        for fd, key in list(sel.get_map().items()):
            # Don't remove our signal handling pipe, control socket, still existing listeners and control connections
            if key.fileobj is signal_r \
                    or (control_socket and key.fileobj is control_socket) \
                    or key.fileobj in listeners \
                    or isinstance(key.fileobj, ControlConnection):
                continue

            # Seems we don't need this one anymore
            sel.unregister(key.fileobj)

        # Collect all the file descriptors we want to listen to
        existing_listeners = [key.fileobj for key in sel.get_map().values()]
        for listener in listeners:
            if listener not in existing_listeners:
                sel.register(listener, selectors.EVENT_READ)

        # Configuration tree
        message_handler = config.create_message_handler()

        # Make sure we have space to store all the interface statistics
        statistics.set_categories(config.statistics)

        # Start worker processes
        with multiprocessing.Pool(processes=config.workers,
                                  initializer=setup_worker,
                                  initargs=(message_handler, logging_queue, lowest_log_level, statistics)) as pool:

            logger.info("Python DHCPv6 server is ready to handle requests")

            running = True
            while running:
                # noinspection PyBroadException
                try:
                    events = sel.select()
                    for key, mask in events:
                        # Handle signal notifications
                        if key.fileobj == signal_r:
                            signal_nr = os.read(signal_r, 1)
                            if signal_nr[0] in (signal.SIGHUP,):
                                # SIGHUP tells the server to reload
                                try:
                                    # Read the new configuration
                                    config = config_parser.load_config(config_file)
                                except (ConfigurationSyntaxError, DataConversionError) as e:
                                    # Make the config exceptions a bit more readable
                                    msg = "Not reloading: " + str(e.message)
                                    if e.lineno and e.lineno != -1:
                                        msg += ' on line {}'.format(e.lineno)
                                    if e.url:
                                        parts = urlparse(e.url)
                                        msg += ' in {}'.format(parts.path)
                                    logger.critical(msg)
                                    continue

                                except ValueError as e:
                                    logger.critical("Not reloading: " + str(e))
                                    continue

                                logger.info("DHCPv6 server restarting after configuration change")
                                running = False
                                stopping = False
                                continue

                            elif signal_nr[0] in (signal.SIGINT, signal.SIGTERM):
                                logger.debug("Received termination request")

                                running = False
                                stopping = True
                                break

                            elif signal_nr[0] in (signal.SIGINFO,):
                                logger.info("Server has processed {} messages".format(message_count))

                        elif isinstance(key.fileobj, ControlSocket):
                            # A new control connection request
                            control_connection = key.fileobj.accept()
                            if control_connection:
                                # We got a connection, listen to events
                                sel.register(control_connection, selectors.EVENT_READ)

                        elif isinstance(key.fileobj, ControlConnection):
                            # Let the connection handle received data
                            control_connection = key.fileobj
                            commands = control_connection.get_commands()
                            for command in commands:
                                if command:
                                    logger.debug("Received control command '{}'".format(command))

                                if command == 'help':
                                    control_connection.send("Recognised commands:")
                                    control_connection.send("  help")
                                    control_connection.send("  stats")
                                    control_connection.send("  stats-json")
                                    control_connection.send("  reload")
                                    control_connection.send("  shutdown")
                                    control_connection.send("  quit")
                                    control_connection.acknowledge()

                                elif command == 'stats':
                                    control_connection.send(str(statistics))
                                    control_connection.acknowledge()

                                elif command == 'stats-json':
                                    control_connection.send(json.dumps(statistics.export()))
                                    control_connection.acknowledge()

                                elif command == 'reload':
                                    # Simulate a SIGHUP to reload
                                    os.write(signal_w, bytes([signal.SIGHUP]))
                                    control_connection.acknowledge('Reloading')

                                elif command == 'shutdown':
                                    # Simulate a SIGTERM to reload
                                    os.write(signal_w, bytes([signal.SIGTERM]))
                                    control_connection.acknowledge('Shutting down')

                                elif command == 'quit' or command is None:
                                    if command == 'quit':
                                        # User nicely signing off
                                        control_connection.acknowledge()

                                    control_connection.close()
                                    sel.unregister(control_connection)
                                    break

                                else:
                                    logger.warning("Rejecting unknown control command '{}'".format(command))
                                    control_connection.reject()

                        elif isinstance(key.fileobj, Listener):
                            packet = key.fileobj.recv_request()

                            # Update stats
                            message_count += 1

                            # Create the callback
                            callback, error_callback = create_handler_callbacks(key.fileobj, packet.message_id)

                            # Dispatch
                            pool.apply_async(handle_message, args=(packet,),
                                             callback=callback, error_callback=error_callback)

                except Exception as e:
                    # Catch-all exception handler
                    logger.exception("Caught unexpected exception {!r}".format(e))

                    now = time.monotonic()

                    # Add new exception time to the history
                    exception_history.append(now)

                    # Remove exceptions outside the window from the history
                    cutoff = now - config.exception_window
                    while exception_history and exception_history[0] < cutoff:
                        exception_history.pop(0)

                    # Did we receive too many exceptions shortly after each other?
                    if len(exception_history) > config.max_exceptions:
                        logger.critical("Received more than {} exceptions in {} seconds, "
                                        "exiting".format(config.max_exceptions, config.exception_window))
                        running = False
                        stopping = True

            pool.close()
            pool.join()

        # Regain root so we can delete the PID file and control socket
        restore_privileges()
        try:
            if pid_filename:
                os.unlink(pid_filename)
                logger.info("Removing PID-file {}".format(pid_filename))
        except OSError:
            pass

        try:
            if control_socket:
                os.unlink(control_socket.socket_path)
                logger.info("Removing control socket {}".format(control_socket.socket_path))
        except OSError:
            pass

    logger.info("Shutting down Python DHCPv6 server v{}".format(dhcpkit.__version__))

    return 0


def run() -> int:
    """
    Run the main program and handle exceptions

    :return: The program exit code
    """
    try:
        # Run the server
        return main(sys.argv[1:])
    except Exception as e:
        logger.critical("Error: {}".format(e))
        return 1


if __name__ == '__main__':
    sys.exit(run())
